"""The ArgusObsTable stores observation information from the Argus survey."""

import logging

import astropy.units as u
import numpy as np
import pandas as pd
from astropy.coordinates import SkyCoord
from cdshealpix import skycoord_to_healpix
from mocpy import MOC

from lightcurvelynx.astro_utils.noise_model import poisson_bandflux_std
from lightcurvelynx.astro_utils.zeropoint import calculate_zp_from_maglim
from lightcurvelynx.consts import GAUSS_EFF_AREA2FWHM_SQ
from lightcurvelynx.obstable.obs_table import ObsTable

_argus_view_radius = 52.0
"""The angular radius of the observation field (in degrees):
https://argus.unc.edu/about

This is not used in search, but provided for the `radius` getter.
"""

_argus_pixel_scale = 1.0
"""The pixel scale for the Argus survey in arcseconds per pixel:
https://argus.unc.edu/specifications
"""


class ArgusHealpixObsTable(ObsTable):
    """An ObsTable for observations from the Argus survey in healpix format.

    Unlike other ObsTable classes, the ArgusHealpixObsTable does not consist of a table of
    pointings, but rather is organized at the healpix level. Each row corresponds
    to a healpix pixel and time.

    Parameters
    ----------
    table : dict or pandas.core.frame.DataFrame
        The table with all the survey information.
    colmap : dict
        A mapping of standard column names to a list of possible names in the input table.
        Each value in the dictionary can be a string or a list of strings.
    saturation_mags : dict, optional
        A dictionary mapping filter names to their saturation thresholds in magnitudes. The filters
        provided must match those in the table. If not provided, Argus-specific defaults will be
        used.
    **kwargs : dict
        Additional keyword arguments to pass to the constructor. This includes overrides
        for survey parameters such as:
        - dark_electrons : The dark current for the camera in electrons per second per pixel.
        - gain: The gain for the camera in electrons per ADU.
        - pixel_scale: The pixel scale for the camera in arcseconds per pixel.
        - radius: The angular radius of the observations (in degrees).
        - read_noise: The readout noise for the camera in electrons per pixel.
    """

    # Column names from the Argus simulation files.
    _argus_sim_colmap = {
        "dark_current": "dark_current",  # The electrons/pixel/exposure
        "dec": "dec",  # degrees
        "exptime": "expTime",  # seconds
        "maglim": "limmag",  # magnitudes
        "ra": "ra",  # degrees
        "seeing": "seeing",  # arcseconds
        "skybrightness": "sky_brightness",  # mag/arcsec^2
        "sky_electrons": "sky_electrons",  # electrons/pixel/exposure
        "time": "epoch",  # MJD
    }
    _default_colnames = _argus_sim_colmap

    # Default survey values: https://argus.unc.edu/specifications
    _default_survey_values = {
        "pixel_scale": _argus_pixel_scale,
        "radius": _argus_view_radius,
        "read_noise": 1.4,  # e-/pixel
        "survey_name": "Argus",
    }

    def __init__(
        self,
        table,
        *,
        colmap=None,
        apply_saturation=True,
        saturation_mags=None,
        nside=None,
        **kwargs,
    ):
        # Set some default values.
        self._spatial_data = None
        self._healpix_nside = nside
        self._healpix_depth = None

        # If the input is a dictionary, convert it to a DataFrame.
        if isinstance(table, dict):
            table = pd.DataFrame(table)

        # The table uses the healpix IDs as the index. We bring this into its own column for easier access
        # and reset the indices.
        if "healpix" in table.columns:
            table = table.reset_index()  # Just reset the index if healpix is already a column.
        elif table.index.name in ["healpix_id", "healpix"]:
            table = table.reset_index().rename(columns={table.index.name: "healpix"})
        else:
            raise ValueError(
                "The input table must have healpix IDs as the index or in a column named 'healpix'."
            )

        # Use the default column mapping if one is not provided.
        if colmap is None:
            colmap = self._default_colnames

        # If filter is not provided, use 'g' for all observations.
        if "filter" not in table.columns:
            table["filter"] = "g"

        # Check the unsupported terms in the kwargs and raise an error if they are provided.
        if "detector_footprint" in kwargs or "wcs" in kwargs:  # pragma: no cover
            raise ValueError("ArgusObsTable does not support detector footprints.")

        super().__init__(
            table=table,
            colmap=colmap,
            apply_saturation=apply_saturation,
            saturation_mags=saturation_mags,
            **kwargs,
        )

    @property
    def healpix_nside(self):
        """Return the nside of the healpix pixels in the table."""
        return self._healpix_nside

    @property
    def healpix_depth(self):
        """Return the depth of the healpix pixels in the table."""
        return self._healpix_depth

    def set_detector_footprint(self, detector_footprint, wcs=None):
        """Set the detector footprint, so footprint filtering is done.

        Parameters
        ----------
        detector_footprint : astropy.regions.SkyRegion, Astropy.regions.PixelRegion, or
            DetectorFootprint
            The footprint object for the instrument's detector.
        wcs : astropy.wcs.WCS, optional
            The WCS for the footprint. Either this or pixel_scale must be provided if
            a footprint is provided as a Astropy region.
        """
        raise NotImplementedError("ArgusObsTable does not support detector footprints.")

    def _build_spatial_data(self):
        """Construct a mapping of healpix id to row number from the ObsTable."""
        if self._healpix_nside is None:
            if "nside" in self._table.columns:
                self._healpix_nside = self._table["nside"][0]
            else:
                raise ValueError(
                    "nside must be provided for ArgusHealpixObsTable construction or "
                    "as a column in the table."
                )

        # Check all nside values are the same.
        if "nside" in self._table.columns:
            nside = self._table["nside"].to_numpy()
            if not np.all(nside == self._healpix_nside):
                raise ValueError(
                    "Inconsistent nside values found in the table. Expected all nside values"
                    f"to be {self._healpix_nside}, but found values: {np.unique(nside)}"
                )

        # Compute the depth.
        self._healpix_depth = int(np.log2(self._healpix_nside))

        # Build a mapping of healpix id to row number.
        self._spatial_data = {}
        index = self._table["healpix"].to_numpy()
        for idx in np.unique(index):
            self._spatial_data[idx] = np.where(index == idx)[0]

    def build_moc(self, max_depth=None, **kwargs):
        """Build a Multi-Order Coverage Map from the regions in the data set.

        These are built directly from the healpix pixels.

        Parameters
        ----------
        max_depth : int, optional
            The maximum depth of the MOC. Default is the depth of the healpix pixels
            in the table.
        **kwargs : dict
            Additional keyword arguments to pass to the MOC construction. Not currently used,
            but accepted for consistency with the ObsTable interface.

        Returns
        -------
        MOC
            The Multi-Order Coverage Map constructed from the data set.
        """
        if max_depth is None:
            max_depth = self._healpix_depth

        logger = logging.getLogger(__name__)
        logger.debug(f"Building MOC from ArgusHealpixObsTable at depth={max_depth}.")
        moc = MOC.from_healpix_cells(
            np.array(list(self._spatial_data.keys())),
            depth=self._healpix_depth,
            max_depth=max_depth,
        )
        return moc

    def _assign_zero_points(self):
        """Assign instrumental zero points in nJy (which produces 1 e-) to the LSSTObsTable tables."""
        if "zp" in self._table.columns:
            return  # Zero points already assigned.

        logger = logging.getLogger(__name__)
        logger.debug("Assigning zero points to the ArgusHealpixObsTable.")

        # Compute the full-width at half-maximum of the PSF in pixels from the
        # seeing (in arcseconds) and the pixel scale (in arcseconds per pixel).
        fwhm_px = self._table["seeing"].to_numpy() / self.safe_get_survey_value("pixel_scale")

        # Compute the dark current in electrons per pixel per second from the dark current in electrons
        # per pixel per exposure and the exposure time.
        exptime = self._table["exptime"].to_numpy()
        dark_current = self._table["dark_electrons"].to_numpy() / exptime

        # Compute the zero points from the 5-sigma depth (and other parmeters).
        zp_vals = calculate_zp_from_maglim(
            maglim=self._table["maglim"],
            sky_bg_electrons=self._table["sky_electrons"],
            fwhm_px=fwhm_px,
            read_noise=self.safe_get_survey_value("read_noise"),
            dark_current=dark_current,
            exptime=exptime,
            nexposure=1,
        )
        self._table["zp"] = zp_vals

    def bandflux_error_point_source(self, bandflux, index):
        """Compute observational bandflux error for a point source

        Parameters
        ----------
        bandflux : array_like of float
            Band bandflux of the point source in nJy.
        index : array_like of int
            The index of the observation in the LSSTObsTable table.

        Returns
        -------
        flux_err : array_like of float
            Simulated bandflux noise in nJy.
        """
        observations = self._table.iloc[index]

        # Compute the PSF footprint in pixels from the seeing and pixel scale.
        pixel_scale = self.safe_get_survey_value("pixel_scale")
        psf_footprint = GAUSS_EFF_AREA2FWHM_SQ * (observations["seeing"] / pixel_scale) ** 2

        # Compute the dark current in electrons per pixel per second from the dark current in electrons
        # per pixel per exposure and the exposure time.
        exptime = observations["exptime"]
        dark_current = observations["dark_electrons"] / exptime

        return poisson_bandflux_std(
            bandflux,
            total_exposure_time=exptime,
            exposure_count=1,
            psf_footprint=psf_footprint,
            sky=observations["sky_electrons"],
            zp=observations["zp"],
            readout_noise=self.safe_get_survey_value("read_noise"),
            dark_current=dark_current,
            zp_err_mag=0.0,  # Placeholder for now
        )

    def range_search(self, query_ra, query_dec, *, radius=None, t_min=None, t_max=None):
        """Return the indices of the pointings that fall within the field
        of view of the query point(s).

        Note that radius is not used since everything is already stored at the healpix level,
        but it is accepted as a parameter for consistency with the ObsTable interface.

        Parameters
        ----------
        query_ra : float or numpy.ndarray
            The query right ascension (in degrees).
        query_dec : float or numpy.ndarray
            The query declination (in degrees).
        radius : float or None, optional
            Not used in this implementation since the data is already organized at the healpix level.
        t_min : float, numpy.ndarray or None, optional
            The minimum time (in MJD) for the observations to consider.
            If None, no time filtering is applied.
        t_max : float, numpy.ndarray or None, optional
            The maximum time (in MJD) for the observations to consider.
            If None, no time filtering is applied.

        Returns
        -------
        inds : list[int] or list[numpy.ndarray]
            Depending on the input, this is either a list of indices for a single query point
            or a list of arrays (of indices) for an array of query points.
        """
        # If the query RA and Dec are scalars, convert them to 1D arrays for consistent processing.
        is_scalar = np.isscalar(query_ra) and np.isscalar(query_dec)
        query_ra = np.atleast_1d(query_ra)
        query_dec = np.atleast_1d(query_dec)
        if len(query_ra) != len(query_dec):
            raise ValueError("Query RA and Dec must have the same length.")

        # Bulk compute the healpix ids for all query points.
        coords = SkyCoord(query_ra * u.deg, query_dec * u.deg, frame="icrs")
        healpix = skycoord_to_healpix(coords, self._healpix_depth)

        # For each query point, get the rows and apply time filtering if specified.
        inds = []
        for hp in healpix:
            if hp in self._spatial_data:
                row_inds = self._spatial_data[hp]

                # Apply time filtering if specified.
                if t_min is not None or t_max is not None:
                    if t_min is not None:
                        times = self._table["time"][row_inds].to_numpy()
                        row_inds = row_inds[times >= t_min]
                    if t_max is not None:
                        times = self._table["time"][row_inds].to_numpy()
                        row_inds = row_inds[times <= t_max]

                inds.append(row_inds)
            else:
                inds.append(np.array([], dtype=int))

        # If the input was a single query point, return a single array of indices instead of a list of arrays.
        if is_scalar:
            inds = inds[0]
        return inds



"""A class for storing and working with simple spectrograph observation tables"""

from __future__ import annotations  # "type1 | type2" syntax in Python <3.10

import numpy as np
import pandas as pd

from lightcurvelynx.obstable.obs_table import ObsTable


class SpectrographObsTable(ObsTable):
    """An ObsTable for simple spectrograph observations.

    Parameters
    ----------
    table : dict or pandas.core.frame.DataFrame
        The table with all the spectrograph observation information.
    colmap : dict
        A mapping of standard column names to a list of possible names in the input table.
        Each value in the dictionary can be a string or a list of strings.
        Defaults to the Rubin CCDVisit column names, stored in _default_colnames.
    **kwargs : dict
        Additional keyword arguments to pass to the constructor.
    """

    _required_names = ["ra", "dec", "time"]

    # By default we do not need any column mapping.
    _default_colnames = {}

    # By default there are no survey values.
    _default_survey_values = {
        "radius": 10.0 / 3600.0,  # degrees, corresponds to a 10 arcsec radius
    }

    # Class constants for the column names.
    def __init__(
        self,
        table,
        colmap=None,
        **kwargs,
    ):
        # If the input table is a dictionary, convert it to a DataFrame.
        if isinstance(table, dict):  # pragma: no cover
            table = pd.DataFrame(table)
        colmap = self._default_colnames if colmap is None else colmap

        # Fill in a dummy filter value if none is given. This will not be used, but
        # is required for the ObsTable constructor.
        if "filter" not in table.columns and (
            "filter" not in colmap or colmap["filter"] not in table.columns
        ):
            table["filter"] = "spectra"  # Name does not actually matter.

        super().__init__(table, colmap=colmap, **kwargs)

    def _assign_zero_points(self):
        """Assign instrumental zero points in nJy (which produces 1 e-) to the LSSTObsTable tables."""
        if "zp" in self._table.columns:
            return  # Nothing to do

        # Add a fake zero point column if we have no information to compute it. This column is not
        # used for spectrographs.
        self.add_column("zp", 0.05 * np.ones(len(self._table)), overwrite=True)



import pickle
import tempfile
from pathlib import Path

import numpy as np
import pytest
from lightcurvelynx.astro_utils.dustmap import ConstantHemisphereDustMap, DustmapWrapper
from lightcurvelynx.effects.extinction import ExtinctionEffect, _DustExtinctionWrapper
from lightcurvelynx.math_nodes.given_sampler import GivenValueList
from lightcurvelynx.models.basic_models import ConstantSEDModel


def test_list_dust_extinction_models():
    """List the available models in the dust_extinction package."""
    model_names = _DustExtinctionWrapper.list_extinction_models()
    assert len(model_names) > 10
    assert "G23" in model_names
    assert "CCM89" in model_names


def test_extinction_effect_dust_extinction():
    """Test that we can create a ExtinctionEffects from the dust_extinction package."""
    for model in ["CCM89", "F99", "G23"]:
        dust_effect = ExtinctionEffect(model, ebv=0.1, r_v=3.1, frame="rest", backend="dust_extinction")

        # We can apply the extinction effect to a set of fluxes.
        fluxes = np.full((10, 3), 1.0)
        wavelengths = np.array([7000.0, 5200.0, 4800.0])
        new_fluxes = dust_effect.apply(fluxes, wavelengths=wavelengths, ebv=0.1, r_v=3.1)
        assert new_fluxes.shape == (10, 3)
        assert np.all(new_fluxes < fluxes)

        # We fail if we are missing a required parameter (either ebv or wavelengths).
        with pytest.raises(ValueError):
            _ = dust_effect.apply(fluxes, wavelengths=wavelengths)
        with pytest.raises(ValueError):
            _ = dust_effect.apply(fluxes, ebv=0.1)

    # We fail if we try to use an invalid model.
    with pytest.raises(KeyError):
        _ = ExtinctionEffect("InvalidModel", ebv=0.1, r_v=3.1, frame="rest", backend="dust_extinction")


def test_extinction_effect_extinction():
    """Test that we can create a ExtinctionEffects from the extinction package."""
    for model in ["odonnell94", "calzetti00"]:
        dust_effect = ExtinctionEffect(model, ebv=0.1, r_v=3.1, frame="rest", backend="extinction")

        # We can apply the extinction effect to a set of fluxes.
        fluxes = np.full((10, 3), 1.0)
        wavelengths = np.array([7000.0, 5200.0, 4800.0])
        new_fluxes = dust_effect.apply(fluxes, wavelengths=wavelengths, ebv=0.1, r_v=3.1)
        assert new_fluxes.shape == (10, 3)
        assert np.all(new_fluxes < fluxes)

        # We fail if we are missing a required parameter (either ebv or wavelengths).
        with pytest.raises(ValueError):
            _ = dust_effect.apply(fluxes, wavelengths=wavelengths)
        with pytest.raises(ValueError):
            _ = dust_effect.apply(fluxes, ebv=0.1)

    # We fail if we try to use an invalid model.
    with pytest.raises(KeyError):
        _ = ExtinctionEffect("InvalidModel", ebv=0.1, r_v=3.1, frame="rest", backend="extinction")

    # We fail for a model that requires r_v if we don't provide it.
    with pytest.raises(ValueError):
        _ = ExtinctionEffect("odonnell94", ebv=0.1, frame="rest", backend="extinction")

    # For the fm07 model r_v is fixed, so we don't need to provide it. But we fail if we
    # try to override it with a different value.
    _ = ExtinctionEffect("fm07", ebv=0.1, frame="rest", backend="extinction")
    with pytest.raises(ValueError):
        _ = ExtinctionEffect("fm07", ebv=0.1, r_v=4.0, frame="rest", backend="extinction")


def test_bad_backend():
    """Test that we raise an error for an invalid backend."""
    with pytest.raises(ValueError):
        _ = ExtinctionEffect("CCM89", ebv=0.1, r_v=3.1, frame="rest", backend="invalid_backend")
    with pytest.raises(ValueError):
        _ = ExtinctionEffect("odonnell94", ebv=0.1, r_v=3.1, frame="rest")


def test_set_frame():
    """Test that correct frame is set"""
    ext = ExtinctionEffect("G23", ebv=0.1, r_v=3.1, frame="observer", backend="dust_extinction")
    assert ext.rest_frame is False

    with pytest.raises(ValueError):
        ExtinctionEffect("G23", ebv=0.1, r_v=3.1, frame="InvalidFrame", backend="dust_extinction")


def test_pickle_extinction_models(subtests):
    """Test that we can pickle and unpickle extinction effects regardless of backend."""
    for backend, model in zip(["dust_extinction", "extinction"], ["F99", "fm07"], strict=False):
        with subtests.test(backend=backend, model=model):
            model = ExtinctionEffect(model, r_v=3.1, frame="rest", ebv=0.1, backend=backend)

            # Compute the some sample fluxes before and after extinction.
            org_fluxes = np.full((10, 3), 1.0)
            wavelengths = np.array([7000.0, 5200.0, 4800.0])
            ext_fluxes_1 = model.apply(org_fluxes, wavelengths=wavelengths, ebv=0.1)
            assert ext_fluxes_1.shape == (10, 3)

            with tempfile.TemporaryDirectory() as tmpdir:
                file_path = Path(tmpdir) / f"test_{backend}_{model}_model.pkl"
                assert not file_path.exists()

                with open(file_path, "wb") as f:
                    pickle.dump(model, f)
                assert file_path.exists()

                with open(file_path, "rb") as f:
                    loaded_model = pickle.load(f)
                assert loaded_model is not None

                ext_fluxes_2 = loaded_model.apply(org_fluxes, wavelengths=wavelengths, ebv=0.1)
                assert ext_fluxes_2.shape == (10, 3)
                assert np.allclose(ext_fluxes_1, ext_fluxes_2)


def test_constant_extinction(subtests):
    """Test that we can create and sample an ExtinctionEffect object."""
    # Use given ebv values. Usually these would be computed from a dustmap,
    # based on (RA, dec).
    ebv_node = GivenValueList([0.1, 0.2, 0.3, 0.1, 0.2, 0.3])
    for backend, ext_name in zip(["dust_extinction", "extinction"], ["F99", "fm07"], strict=False):
        with subtests.test(backend=backend, model=ext_name):
            dust_effect = ExtinctionEffect(ext_name, r_v=3.1, frame="rest", ebv=ebv_node, backend=backend)

            model = ConstantSEDModel(
                brightness=100.0,
                ra=0.0,
                dec=40.0,
                redshift=0.0,
            )
            model.add_effect(dust_effect)

            times = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
            wavelengths = np.array([7000.0, 5200.0, 4800.0])  # Red, green, blue
            states = model.sample_parameters(num_samples=3)
            fluxes = model.evaluate_sed(times, wavelengths, states)

            # We check that all fluxes are reduced, and that higher ebv leads to
            # lower fluxes.
            assert fluxes.shape == (3, 5, 3)
            assert np.all(fluxes < 100.0)
            assert np.all(fluxes[0, :, :] > fluxes[1, :, :])
            assert np.all(fluxes[1, :, :] > fluxes[2, :, :])


def test_dustmap_chain(subtests):
    """Test that we can chain the dustmap computation and extinction effect."""
    for backend, ext_name in zip(["dust_extinction", "extinction"], ["F99", "fm07"], strict=False):
        with subtests.test(backend=backend, model=ext_name):
            model = ConstantSEDModel(
                brightness=100.0,
                ra=GivenValueList([45.0, 45.0, 45.0, 45.0, 45.0, 45.0]),
                dec=GivenValueList([20.0, -20.0, 10.0, 20.0, -20.0, 10.0]),
                redshift=0.0,
            )

            # Create a constant dust map for testing.
            dust_map = ConstantHemisphereDustMap(north_ebv=0.8, south_ebv=0.5)
            dust_map_node = DustmapWrapper(dust_map, ra=model.ra, dec=model.dec)

            # Create an extinction effect using the EBVs from that dust map.
            ext_effect = ExtinctionEffect(
                extinction_model=ext_name,
                ebv=dust_map_node,
                r_v=3.1,
                frame="rest",
                backend=backend,
            )
            model.add_effect(ext_effect)

            # Sample the model.
            times = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
            wavelengths = np.array([7000.0, 5200.0])
            states = model.sample_parameters(num_samples=3)
            fluxes = model.evaluate_sed(times, wavelengths, states)

            assert fluxes.shape == (3, 5, 2)
            assert np.all(fluxes < 100.0)
            assert np.allclose(fluxes[0, :, :], fluxes[2, :, :])
            assert np.all(fluxes[1, :, :] > fluxes[0, :, :])



"""A class for storing and working with Rubin Observatory (LSST) observation tables."""

from __future__ import annotations  # "type1 | type2" syntax in Python <3.10

import numpy as np

from lightcurvelynx.astro_utils.detector_footprint import DetectorFootprint
from lightcurvelynx.astro_utils.mag_flux import mag2flux
from lightcurvelynx.astro_utils.noise_model import poisson_bandflux_std
from lightcurvelynx.consts import GAUSS_EFF_AREA2FWHM_SQ
from lightcurvelynx.obstable.obs_table import ObsTable

LSSTCAM_PIXEL_SCALE = 0.2
"""The pixel scale for the LSST camera in arcseconds per pixel."""

_lsstcam_readout_noise = 5.82
"""The standard deviation of the count of readout electrons per pixel for the LSST
camera. This is the average value from the two CCD types used (e2v=5.40 and ITL=6.21)
in the LSST camera.

Averaging as sqrt((5.40^2 + 6.21^2) / 2) = 5.82

The value is from https://lsstcam.lsst.io/index.html
"""

_lsstcam_dark_current = 0.022
"""The dark current for the LSST camera in electrons per second per pixel.
This is the average value from the two CCD types used (e2v=0.023 and ITL=0.021)
in the LSST camera.

The value is from https://lsstcam.lsst.io/index.html
"""

_lsstcam_gain = 1.595
"""The gain for the LSST camera in electrons per ADU.
This is the average value from the two CCD types used (e2v=1.51 and ITL=1.68)
in the LSST camera.

The value is from https://lsstcam.lsst.io/index.html
"""

_lsstcam_view_radius = 1.75
"""The angular radius of the observation field (in degrees). LSSTCam field of view is 9.6 square degree """

_lsstcam_ccd_radius = 0.1574
"""The approximate angular radius of a single LSST CCD (in degrees). Each CCD is 800*800 arcsec^2.
We approximate the radius as 800 arcsec/ sqrt(2). We overestimate slightly, because this value is
used in range searches. More exact filtering is done with the detector footprint.
"""

_lsst_zp_err_mag = 1.0e-4
"""The zero point error in magnitude.

We choose a very conservative noise flooring of 1e-4 mag.
This number will be updated when we have a better estimate from LSST.
"""


class LSSTObsTable(ObsTable):
    """An ObsTable for observations from the Rubin Observatory data releases.

    Parameters
    ----------
    table : dict or pandas.core.frame.DataFrame
        The table with all the LSST survey information.
    colmap : dict
        A mapping of standard column names to a list of possible names in the input table.
        Each value in the dictionary can be a string or a list of strings.
        Defaults to the Rubin CCDVisit column names, stored in _default_colnames.
    saturation_mags : dict, optional
        A dictionary mapping filter names to their saturation thresholds in magnitudes. The filters
        provided must match those in the table. If not provided, LSST-specific defaults will be
        used.
    **kwargs : dict
        Additional keyword arguments to pass to the constructor. This includes overrides
        for survey parameters such as:
        - dark_current : The dark current for the camera in electrons per second per pixel.
        - gain: The gain for the camera in electrons per ADU.
        - pixel_scale: The pixel scale for the camera in arcseconds per pixel.
        - radius: The angular radius of the observations (in degrees).
        - read_noise: The readout noise for the camera in electrons per pixel.
    """

    _required_names = ["ra", "dec", "time"]

    # Column names for the Rubin CCDVisit table from schemas including
    #   * DP1 (https://sdm-schemas.lsst.io/dp1.html#CcdVisit)
    #   * DP2+ (https://sdm-schemas.lsst.io/lsstcam.html#CcdVisit)
    _ccdvisit_colmap = {
        "dec": "dec",  # degrees
        "exptime": "expTime",  # seconds
        "filter": "band",
        "maglim": "magLim",  # magnitudes
        "pixel_scale": "pixelScale",  # arcseconds per pixel
        "ra": "ra",  # degrees
        "rotation": "skyRotation",  # degrees
        "seeing": "seeing",  # arcseconds
        "sky_bg_adu": "skyBg",  # Averge sky background in ADU
        "time": ["expMidptMJD", "obsStartMJD"],  # days
        "zp_mag_adu": "zeroPoint",  # magnitudes to produce 1 count (ADU)
    }

    # Column names for the Rubin Science Validation visit data release, which has
    # a different schema than the DP1 and DP2+ CCDVisit tables (mostly similar to OpSim):
    # https://survey-strategy.lsst.io/progress/sv_status/sv_20250930.html
    # See also:
    # https://github.com/lsst/tutorial-notebooks/blob/main/Commissioning/101_lsstcam_visits_database.ipynb
    _sv_visits_colmap = {
        # Values defined in the OpSim schema:
        # https://rubin-scheduler.lsst.io/fbs-output-schema.html
        "dec": "fieldDec",  # degrees
        "exptime": "visitExposureTime",  # seconds
        "filter": "band",
        "maglim": "fiveSigmaDepth",  # magnitudes
        "ra": "fieldRA",  # degrees
        "rotation": "rotSkyPos",  # degrees
        "seeing": "seeingFwhmEff",  # arcseconds
        "time": "exp_midpt_mjd",  # days
        # Some of the values are defined in the ConsDB schema:
        # https://sdm-schemas.lsst.io/cdb_lsstcam.html#exposure
        "sky_bg_e": "sky_bg_median",  # Averge sky background in electrons per pixel
        "zp_mag_e": "zero_point_median",  # magnitude to produce 1 electron per exposure
    }

    # For now use the CCDVisit column mapping as the default.
    _default_colnames = _ccdvisit_colmap

    # Default survey values (LSSTCam).
    _default_survey_values = {
        "ccd_pixel_width": 4000,
        "ccd_pixel_height": 4000,
        "dark_current": _lsstcam_dark_current,
        "gain": _lsstcam_gain,
        "pixel_scale": LSSTCAM_PIXEL_SCALE,
        "radius": _lsstcam_view_radius,
        "read_noise": _lsstcam_readout_noise,
        "zp_err_mag": _lsst_zp_err_mag,
        "survey_name": "LSST",
    }

    # Default LSST saturation thresholds in magnitudes.
    # https://www.lsst.org/sites/default/files/docs/sciencebook/SB_3.pdf
    _default_saturation_mags = {
        "u": 14.7,
        "g": 15.7,
        "r": 15.8,
        "i": 15.8,
        "z": 15.3,
        "y": 13.9,
    }

    # Class constants for the column names.
    def __init__(
        self,
        table,
        colmap=None,
        saturation_mags=None,
        **kwargs,
    ):
        colmap = self._default_colnames if colmap is None else colmap

        # If saturation thresholds are not provided, then set to the
        # LSSTObsTable defaults.
        if saturation_mags is None:
            saturation_mags = self._default_saturation_mags

        super().__init__(table, colmap=colmap, saturation_mags=saturation_mags, **kwargs)

    def _assign_zero_points(self):
        """Assign instrumental zero points in nJy (which produces 1 e-) to the LSSTObsTable tables."""
        cols = self._table.columns.to_list()

        if "zp" in cols:
            return  # Nothing to do

        # If the zero point column is already present (as a magnitude),
        # we convert it to nJy.
        if "zp_mag_adu" in cols:
            zp_values = mag2flux(self._table["zp_mag_adu"]) / self.safe_get_survey_value("gain")
            self.add_column("zp", zp_values, overwrite=True)
            return
        if "zp_mag_e" in cols:
            zp_values = mag2flux(self._table["zp_mag_e"])
            self.add_column("zp", zp_values, overwrite=True)
            return

        raise ValueError("Not enough information to compute the zero points.")

    @classmethod
    def from_ccdvisit_table(cls, table, make_detector_footprint=False, **kwargs):
        """Construct an LSSTObsTable object from a CCDVisit table.

        As an example we could access the DP1 CCDVisit table from RSP as:
            from lsst.rsp import get_tap_service
            service = get_tap_service("tap")
            table = service.search("SELECT * FROM dp1.CcdVisit").to_table().to_pandas()

        Or you can read a table from a file (e.g. using the `read_sqlite_table` function).
            from lightcurvelynx.utils.io_utils import read_sqlite_table
            table = read_sqlite_table("path_to_file.db", sql_query="SELECT * FROM observations")

        Parameters
        ----------
        table : pandas.core.frame.DataFrame
            The CCDVisit table containing the LSSTObsTable data.
        make_detector_footprint : bool, optional
            If True, the detector footprint will be created based on the xSize and ySize columns
            in the table.
        **kwargs : dict
            Additional keyword arguments to pass to the LSSTObsTable constructor.

        Returns
        -------
        obstable : LSSTObsTable
            An LSSTObsTable object containing the data from the CCDVisit table.
        """
        table = table.copy()
        cols = table.columns.to_list()

        # Try to derive the viewing radius if we have the information to do so.
        if "xSize" in cols and "ySize" in cols and "pixel_scale" in cols:
            radius_px = np.sqrt((table["xSize"] / 2) ** 2 + (table["ySize"] / 2) ** 2)
            table["radius"] = (radius_px * table["pixel_scale"]) / 3600.0  # arcsec to degrees
        elif "radius" not in kwargs:
            # Use a single approximate average ccd radius.
            kwargs["radius"] = _lsstcam_ccd_radius

        # Create the ObsTable object. Use the default column mapping for LSST, which
        # supports the DP1 and DP2+ CCDVisit Table schemas.
        obstable = cls(table, **kwargs)

        # Create a detector footprint if requested. We use the same (average) footprint for
        #  all CCDs based on the survey parameters for pixel scale and CCD size.
        if make_detector_footprint:
            pixel_scale = obstable.survey_values.get("pixel_scale")
            width_px = obstable.survey_values.get("ccd_pixel_width")
            height_px = obstable.survey_values.get("ccd_pixel_height")
            detect_fp = DetectorFootprint.from_pixel_rect(width_px, height_px, pixel_scale=pixel_scale)
            obstable.set_detector_footprint(detect_fp)

        return obstable

    @classmethod
    def from_sv_visits_table(cls, table, **kwargs):
        """Construct an LSSTObsTable object from a science validation visits table.
        https://survey-strategy.lsst.io/progress/sv_status/sv_20250930.html
        Note this table uses a combination of the schemas (e.g. OpSim and ConsDB).

        As an example we can read a table from a file (e.g. using the `read_sqlite_table` function).
            from lightcurvelynx.utils.io_utils import read_sqlite_table
            table = read_sqlite_table("path_to_file.db", sql_query="SELECT * FROM observations")

        Parameters
        ----------
        table : pandas.core.frame.DataFrame
            The science validation visits table containing the LSSTObsTable data.
        **kwargs : dict
            Additional keyword arguments to pass to the LSSTObsTable constructor.

        Returns
        -------
        obstable : LSSTObsTable
            An LSSTObsTable object containing the data from the science validation visits table.
        """
        # Remove the rows with NaNs in sky_bg_median or zero_point_median.
        table = table.dropna(subset=["sky_bg_median", "zero_point_median"])

        # Set the radius as the view (not CCD) radius because we do not have
        # per-CCD information.
        if "radius" not in kwargs:
            kwargs["radius"] = _lsstcam_view_radius

        # Create the ObsTable object using the science validation visits column mapping
        # (if a custom column mapping is not provided).
        colmap = kwargs.pop("colmap", cls._sv_visits_colmap)
        obstable = cls(table, colmap=colmap, **kwargs)
        return obstable

    def bandflux_error_point_source(self, bandflux, index):
        """Compute observational bandflux error for a point source

        Parameters
        ----------
        bandflux : array_like of float
            Band bandflux of the point source in nJy.
        index : array_like of int
            The index of the observation in the LSSTObsTable table.

        Returns
        -------
        flux_err : array_like of float
            Simulated bandflux noise in nJy.
        """
        observations = self._table.iloc[index]

        # By the effective FWHM definition, see
        # https://smtn-002.lsst.io/v/OPSIM-1171/index.html
        # We need it in pixel^2
        pixel_scale = self.safe_get_survey_value("pixel_scale")
        psf_footprint = GAUSS_EFF_AREA2FWHM_SQ * (observations["seeing"] / pixel_scale) ** 2
        zp = observations["zp"]

        # Compute sky background in e- from sky_bg_adu.
        if "sky_bg_e" in observations:
            sky = observations["sky_bg_e"]
        elif "sky_bg_adu" in observations:
            sky = observations["sky_bg_adu"] * self.safe_get_survey_value("gain")
        else:
            raise ValueError(
                "Some value of sky background (electrons or ADU) is required to compute "
                "the sky background noise."
            )

        # If exposure time is not provided, use the default value of 30 seconds for LSST.
        exptime = 30.0 if "exptime" not in observations.columns else observations["exptime"]

        return poisson_bandflux_std(
            bandflux,
            total_exposure_time=exptime,
            exposure_count=1,
            psf_footprint=psf_footprint,
            sky=sky,
            zp=zp,
            readout_noise=self.safe_get_survey_value("read_noise"),
            dark_current=self.safe_get_survey_value("dark_current"),
            zp_err_mag=self.safe_get_survey_value("zp_err_mag"),
        )



import matplotlib

matplotlib.use("Agg")  # Suppress the plots

import tempfile
import warnings
from pathlib import Path

import astropy.units as u
import numpy as np
import pandas as pd
import pytest
from astropy.coordinates import Latitude, Longitude, SkyCoord, angular_separation
from lightcurvelynx.astro_utils.detector_footprint import DetectorFootprint
from lightcurvelynx.math_nodes.ra_dec_sampler import (
    ApproximateMOCSampler,
    ObsTableRADECSampler,
    ObsTableUniformRADECSampler,
    UniformRADEC,
)
from lightcurvelynx.math_nodes.single_value_node import SingleVariableNode
from lightcurvelynx.obstable.opsim import OpSim
from lightcurvelynx.utils.io_utils import write_results_as_hats
from mocpy import MOC
from nested_pandas import NestedFrame


def test_uniform_ra_dec():
    """Test that we can generate numbers from a uniform distribution on a sphere."""
    sampler_node = UniformRADEC(seed=100, node_label="sampler")

    # Test we can generate a single value.
    (ra, dec) = sampler_node.generate(num_samples=1)
    assert 0.0 <= ra <= 360.0
    assert -90.0 <= dec <= 90.0

    # Generate many samples.
    num_samples = 20_000
    state = sampler_node.sample_parameters(num_samples=num_samples)

    all_ra = state["sampler"]["ra"]
    assert len(all_ra) == num_samples
    assert np.all(all_ra >= 0.0)
    assert np.all(all_ra <= 360.0)

    all_dec = state["sampler"]["dec"]
    assert len(all_dec) == num_samples
    assert np.all(all_dec >= -90.0)
    assert np.all(all_dec <= 90.0)

    # Compute histograms of RA and dec values.
    ra_bins = np.zeros(36)
    dec_bins = np.zeros(18)
    for idx in range(num_samples):
        ra_bins[int(all_ra[idx] / 10.0)] += 1
        dec_bins[int((all_dec[idx] + 90.0) / 10.0)] += 1

    # Check that all RA bins have approximately equal samples.
    expected_count = num_samples / 36
    for bin_count in ra_bins:
        assert 0.8 <= bin_count / expected_count <= 1.2

    # Check that the dec bins around the poles have less samples
    # than the bins around the equator.
    assert dec_bins[0] < 0.25 * dec_bins[9]
    assert dec_bins[17] < 0.25 * dec_bins[10]

    # Check that we can generate uniform samples in radians.
    sampler_node2 = UniformRADEC(seed=100, node_label="sampler2", use_degrees=False)
    state2 = sampler_node2.sample_parameters(num_samples=num_samples)

    all_ra = state2["sampler2"]["ra"]
    assert len(all_ra) == num_samples
    assert np.all(all_ra >= 0.0)
    assert np.all(all_ra <= 2.0 * np.pi)

    all_dec = state2["sampler2"]["dec"]
    assert len(all_dec) == num_samples
    assert np.all(all_dec >= -np.pi)
    assert np.all(all_dec <= np.pi)


def test_obstable_ra_dec_sampler():
    """Test that we can sample from am OpSim object."""
    values = {
        "observationStartMJD": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "fieldRA": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "fieldDec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp_nJy": np.ones(5),
    }
    ops_data = OpSim(values)
    assert len(ops_data) == 5

    sampler_node = ObsTableRADECSampler(ops_data, radius=0.0, node_label="sampler")
    assert sampler_node.radius == 0.0

    # Test we can generate a single random value and the entries
    # are consistent across the row.
    state = sampler_node.sample_parameters(num_samples=1)
    idx = state["sampler"]["selected_table_index"]
    assert state["sampler"]["ra"] == values["fieldRA"][idx]
    assert state["sampler"]["dec"] == values["fieldDec"][idx]
    assert state["sampler"]["time"] == values["observationStartMJD"][idx]

    # Check that we returned floats, not arrays.
    assert isinstance(state["sampler"]["ra"], float)
    assert isinstance(state["sampler"]["dec"], float)
    assert isinstance(state["sampler"]["time"], float)

    # Do randomized sampling (with no offset).
    sampler_node2 = ObsTableRADECSampler(ops_data, radius=0.0, seed=100, node_label="sampler")
    state = sampler_node2.sample_parameters(num_samples=5000)

    # Check that the samples are uniform and consistent.
    int_times = state["sampler"]["time"].astype(int)
    assert np.allclose(state["sampler"]["ra"], values["fieldRA"][int_times])
    assert np.allclose(state["sampler"]["dec"], values["fieldDec"][int_times])
    assert np.array_equal(state["sampler"]["selected_table_index"], int_times)
    assert len(int_times[int_times == 0]) > 750
    assert len(int_times[int_times == 1]) > 750
    assert len(int_times[int_times == 2]) > 750
    assert len(int_times[int_times == 3]) > 750
    assert len(int_times[int_times == 4]) > 750

    # Do randomized sampling with offsets (using the default OpSim radius).
    sampler_node3 = ObsTableRADECSampler(ops_data, seed=100, node_label="sampler")
    state = sampler_node3.sample_parameters(num_samples=5000)

    # Check that the samples are not all the centers (unique values > 500) but are within
    # the sampling radius of the corresponding pointing. We use int(time) to match the index
    # of the center pointing.
    int_times = state["sampler"]["time"].astype(int)
    assert len(np.unique(state["sampler"]["ra"])) > 500
    assert len(np.unique(state["sampler"]["dec"])) > 500
    for idx in range(5):
        dist = angular_separation(
            state["sampler"]["ra"][int_times == idx] * u.deg,
            state["sampler"]["dec"][int_times == idx] * u.deg,
            values["fieldRA"][idx] * u.deg,
            values["fieldDec"][idx] * u.deg,
        )
        assert np.all(dist <= sampler_node3.radius * u.deg)

    # We fail if no radius is provided by the OpSim or the parameters
    ops_data = OpSim(values, radius=None)
    with pytest.raises(ValueError):
        _ = ObsTableRADECSampler(ops_data)

    # But we can use the OpSim's radius if that is provided.
    ops_data = OpSim(values, radius=1.0)
    sampler_node = ObsTableRADECSampler(ops_data)
    assert sampler_node.radius == 1.0


def test_obstable_ra_dec_sampler_extra():
    """Test that we can sample from an ObsTable-like object with extra columns."""
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp": np.ones(5),
        "extra": np.array([10, 20, 30, 40, 50]),
    }
    ops_df = pd.DataFrame(values)
    sampler_node = ObsTableRADECSampler(ops_df, radius=0.0, extra_cols=["extra"], node_label="sampler")
    assert sampler_node.radius == 0.0

    # Test we can generate a single value.
    sample = sampler_node.sample_parameters(num_samples=1)
    idx = sample["sampler"]["selected_table_index"]
    assert sample["sampler"]["ra"] == values["ra"][idx]
    assert sample["sampler"]["dec"] == values["dec"][idx]
    assert sample["sampler"]["time"] == values["time"][idx]
    assert sample["sampler"]["extra"] == values["extra"][idx]

    # We can chain on any of the column names.
    single_node = SingleVariableNode("extra", sampler_node.extra, node_label="single")
    state = single_node.sample_parameters(num_samples=2)
    assert state["single.extra"][0] in values["extra"]
    assert state["single.extra"][1] in values["extra"]


def test_obstable_ra_dec_sampler_from_hats(test_data_dir):
    """Test that we can sample from a HATS catalog on disk."""
    outer_dict = {
        "id": [0, 1, 2],
        "ra": [10.0, 10.1, 10.2],
        "dec": [-10.0, -9.9, -10.1],
        "nobs": [3, 2, 1],
        "z": [0.1, 0.2, 0.3],
    }
    inner_dict = {
        "mjd": [59000, 59001, 59002, 59000, 59001, 59000],
        "flux": [10.0, 12.0, 11.0, 15.0, 14.0, 13.0],
        "fluxerr": [1.0, 1.0, 1.0, 1.5, 1.5, 1.0],
        "filter": ["g", "r", "i", "g", "r", "i"],
    }
    nested_inds = [0, 0, 0, 1, 1, 2]
    results = NestedFrame(data=outer_dict, index=[0, 1, 2])
    nested_1 = pd.DataFrame(data=inner_dict, index=nested_inds)
    results = results.join_nested(nested_1, "lightcurve")

    with tempfile.TemporaryDirectory() as dir_name:
        dir_path = Path(dir_name, "test_hats")
        write_results_as_hats(dir_path, results)

        sampler_node = ObsTableRADECSampler.from_hats(
            dir_path,
            radius=0.0,
            node_label="sampler",
            extra_cols=["z"],
        )
        assert sampler_node.radius == 0.0

        states = sampler_node.sample_parameters(num_samples=3)
        for i in range(3):
            idx = states["sampler"]["selected_table_index"][i]
            assert states["sampler"]["ra"][i] == outer_dict["ra"][idx]
            assert states["sampler"]["dec"][i] == outer_dict["dec"][idx]
            assert states["sampler"]["z"][i] == outer_dict["z"][idx]


def test_opsim_uniform_ra_dec_sampler():
    """Test that we can sample uniformly from am OpSim object."""
    # Create an opsim with two points in different hemispheres.
    values = {
        "observationStartMJD": np.array([0.0, 1.0]),
        "fieldRA": np.array([15.0, 195.0]),
        "fieldDec": np.array([75.0, -75.0]),
        "zp_nJy": np.ones(2),
    }
    ops_data = OpSim(values)
    assert len(ops_data) == 2

    # Use a very large radius so we do not reject too many samples.
    sampler_node = ObsTableUniformRADECSampler(ops_data, radius=30.0, seed=100, node_label="sampler")
    assert sampler_node.radius == 30.0

    # Test we can generate a single value.
    ra, dec = sampler_node.generate(num_samples=1)
    assert ops_data.is_observed(ra, dec, radius=30.0)

    # Test we can generate many observations
    num_samples = 10_000
    ra, dec = sampler_node.generate(num_samples=num_samples)
    assert np.all(ops_data.is_observed(ra, dec, radius=30.0))

    # We should sample roughly uniformly from the two regions.
    northern_mask = dec > 0.0
    assert np.sum(northern_mask) > 0.4 * num_samples
    assert np.sum(northern_mask) < 0.6 * num_samples

    northern_center = SkyCoord(ra=15.0, dec=75.0, unit="deg")
    northern_pts = SkyCoord(ra[northern_mask], dec[northern_mask], unit="deg")
    assert np.all(northern_pts.separation(northern_center) <= 30.0 * u.deg)

    southern_center = SkyCoord(ra=195.0, dec=-75.0, unit="deg")
    southern_pts = SkyCoord(ra[~northern_mask], dec[~northern_mask], unit="deg")
    assert np.all(southern_pts.separation(southern_center) <= 30.0 * u.deg)

    # We fail if neither the OpSim or the sampler have a radius
    ops_data = OpSim(values, radius=None)
    with pytest.raises(ValueError):
        _ = ObsTableUniformRADECSampler(ops_data)

    # But we succeed if the OpSim has a radius.
    ops_data = OpSim(values, radius=10.0)
    sampler_node = ObsTableUniformRADECSampler(ops_data)
    assert sampler_node.radius == 10.0


def test_opsim_uniform_ra_dec_sampler_footprint():
    """Test that we can sample uniformly from am OpSim object with a footprint."""
    # Create an opsim with two points in different hemispheres.
    values = {
        "observationStartMJD": np.array([0.0, 1.0]),
        "fieldRA": np.array([15.0, 195.0]),
        "fieldDec": np.array([75.0, -75.0]),
        "zp_nJy": np.ones(2),
    }

    # Detector is a rectangle of 20 x 30 degrees.
    fp = DetectorFootprint.from_sky_rect(width=20, height=30, pixel_scale=0.1)
    ops_data = OpSim(values, detector_footprint=fp)

    # We fail if we try to use negative number of iterations.
    with pytest.raises(ValueError):
        _ = ObsTableUniformRADECSampler(ops_data, radius=30.0, max_iterations=-1)

    # Test we can generate multiple observations
    num_samples = 200
    sampler_node = ObsTableUniformRADECSampler(ops_data, radius=50.0, seed=100, node_label="sampler")
    ra, dec = sampler_node.generate(num_samples=num_samples)

    # We should sample roughly uniformly from the two regions.
    northern_mask = dec > 0.0
    assert np.sum(northern_mask) > 0.4 * num_samples
    assert np.sum(northern_mask) < 0.6 * num_samples

    # All the points should be visible.
    assert np.all(ops_data.is_observed(ra, dec, radius=50.0))


def test_approximate_moc_sampler():
    """Test that we can create and sample from an ApproximateMOCSampler."""
    longitudes = Longitude([15.0, 90.0], unit="deg")
    latitudes = Latitude([-20.0, 20.0], unit="deg")
    moc = MOC.from_cones(
        lon=longitudes,
        lat=latitudes,
        radius=1.0 * u.deg,
        max_depth=12,
        union_strategy="large_cones",
    )
    moc_sampler = ApproximateMOCSampler(moc)

    # Test we can generate a single value.
    ra, dec = moc_sampler.generate(num_samples=1)
    assert moc.contains_skycoords(SkyCoord(ra, dec, unit="deg"))

    # Test we can generate many observations
    num_samples = 10_000
    ra, dec = moc_sampler.generate(num_samples=num_samples)
    assert np.all(moc.contains_skycoords(SkyCoord(ra, dec, unit="deg")))

    # We should sample roughly uniformly from the two regions.
    northern_mask = dec > 0.0
    assert np.sum(northern_mask) > 0.4 * num_samples
    assert np.sum(northern_mask) < 0.6 * num_samples

    assert np.all(ra[northern_mask] > 88.0)
    assert np.all(ra[northern_mask] < 92.0)
    assert np.all(ra[~northern_mask] > 13.0)
    assert np.all(ra[~northern_mask] < 17.0)
    assert np.all(dec[northern_mask] > 18.0)
    assert np.all(dec[northern_mask] < 22.0)
    assert np.all(dec[~northern_mask] > -22.0)
    assert np.all(dec[~northern_mask] < -18.0)

    # We fail with an invalid depth.
    with pytest.raises(ValueError):
        ApproximateMOCSampler(moc, depth=2000)


def test_approximate_moc_sampler_plot():
    """Test that we can plot the footprint of an ApproximateMOCSampler."""
    longitudes = Longitude([15.0, 90.0], unit="deg")
    latitudes = Latitude([-20.0, 20.0], unit="deg")
    moc = MOC.from_cones(
        lon=longitudes,
        lat=latitudes,
        radius=1.0 * u.deg,
        max_depth=12,
        union_strategy="large_cones",
    )
    moc_sampler = ApproximateMOCSampler(moc)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        moc_sampler.plot_footprint(depth=12)
        matplotlib.pyplot.close("all")


def test_approximate_moc_sampler_from_file():
    """Test that we can create an ApproximateMOCSampler from a MOC file."""
    longitudes = Longitude([15.0, 90.0], unit="deg")
    latitudes = Latitude([-20.0, 20.0], unit="deg")
    moc = MOC.from_cones(
        lon=longitudes,
        lat=latitudes,
        radius=1.0 * u.deg,
        max_depth=8,
        union_strategy="large_cones",
    )

    with tempfile.TemporaryDirectory() as dir_name:
        for fmt in ["fits", "json", "ascii"]:
            file_path = Path(dir_name, f"test_mock.{fmt}")

            # We can't load a non-existent file.
            with pytest.raises(FileNotFoundError):
                _ = ApproximateMOCSampler.from_file(file_path, format=fmt)

            moc.save(file_path, format=fmt)
            moc_sampler = ApproximateMOCSampler.from_file(file_path, format=fmt)

            assert moc_sampler.healpix_list is not None


def test_approximate_moc_sampler_from_obstable():
    """Test that we can create an ApproximateMOCSampler from an ObsTable."""
    values = {
        "observationStartMJD": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "fieldRA": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "fieldDec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp_nJy": np.ones(5),
    }
    ops_data = OpSim(values)
    assert len(ops_data) == 5

    # Create the MOC sampler and test that we can generate samples that are "near"
    # the given centers. We add a bit of extra radius to account for the approximate nature
    # of the MOC at depth=12.
    moc_sampler = ApproximateMOCSampler.from_obstable(ops_data, depth=12, radius=1.75, seed=43)
    ra, dec = moc_sampler.generate(num_samples=100)
    assert np.all(ops_data.is_observed(ra, dec, radius=1.8))


def test_approximate_moc_sampler_from_union_obstable():
    """Test that we can create an ApproximateMOCSampler from a union of ObsTables."""
    # Create four ObsTables with (mostly) different pointings. The 4th is a copy of the 1st.
    ops_tables = []
    for center_ra, center_dec in [(15.0, -10.0), (30.0, -5.0), (15.0, 0.0), (15.0, -10.0)]:
        values = {
            # Each survey is 3 pointings on the same declination.
            "observationStartMJD": np.array([0.0, 1.0, 2.0]),
            "fieldRA": np.array([center_ra, center_ra - 0.1, center_ra + 0.1]),
            "fieldDec": np.array([center_dec, center_dec, center_dec]),
            "zp_nJy": np.ones(3),
        }
        ops_data = OpSim(values)
        ops_tables.append(ops_data)

    # Build a single sampler from the union of the ObsTables.
    moc_sampler = ApproximateMOCSampler.from_obstable_union(
        ops_tables,
        depth=14,
        radii=[0.2, 0.2, 0.2, 0.2],
        seed=43,
    )

    # Generate 1000 samples.
    ra_arr, dec_arr = moc_sampler.generate(num_samples=1_000)
    sample_coords = SkyCoord(ra=ra_arr * u.deg, dec=dec_arr * u.deg)

    close_mask = np.zeros(len(ra_arr), dtype=bool)
    for center_ra, center_dec in [(15.0, -10.0), (30.0, -5.0), (15.0, 0.0)]:
        for ra_offset in [-0.1, 0.0, 0.1]:
            center_coord = SkyCoord(ra=(center_ra + ra_offset) * u.deg, dec=center_dec * u.deg)
            separations = center_coord.separation(sample_coords)
            is_close = separations <= 0.25 * u.deg  # Small buffer for MOC approximation
            close_mask |= is_close

            # Check we sampled at least one point from this pointing.
            assert np.any(is_close)

    # Check that all points are close to something.
    assert np.all(close_mask)

    # We can create a sampler with no radii (just using the default for each table).
    moc_sampler2 = ApproximateMOCSampler.from_obstable_union(ops_tables, depth=10)
    assert moc_sampler2 is not None

    # We fail if the list is empty.
    with pytest.raises(ValueError):
        _ = ApproximateMOCSampler.from_obstable_union([], depth=10, radii=[], seed=45)

    # We fail if the list lengths do not match.
    with pytest.raises(ValueError):
        _ = ApproximateMOCSampler.from_obstable_union(ops_tables, depth=10, radii=[1.0], seed=45)


def test_approximate_moc_sampler_from_intersection_obstable():
    """Test that we can create an ApproximateMOCSampler from a union of ObsTables."""
    center_coords = [(15.0, -10.0), (16.0, -10.0), (15.5, -9.0)]

    # Create several ObsTables with different, but close, pointings.
    ops_tables = []
    for center_ra, center_dec in center_coords:
        values = {
            # Each survey is 3 pointings on the same declination.
            "observationStartMJD": np.array([0.0]),
            "fieldRA": np.array([center_ra]),
            "fieldDec": np.array([center_dec]),
            "zp_nJy": np.ones(1),
        }
        ops_data = OpSim(values)
        ops_tables.append(ops_data)

    # Build a single sampler from the union of the ObsTables.
    moc_sampler = ApproximateMOCSampler.from_obstable_intersection(
        ops_tables,
        depth=14,
        radii=[1.0, 1.0, 1.0],
        seed=45,
    )

    # Generate 1000 samples.
    ra_arr, dec_arr = moc_sampler.generate(num_samples=1_000)
    sample_coords = SkyCoord(ra=ra_arr * u.deg, dec=dec_arr * u.deg)

    # Check that all of the points are in to the overlapping region (within
    # 1 degree of all the centers).
    for center_ra, center_dec in center_coords:
        center_coord = SkyCoord(ra=center_ra * u.deg, dec=center_dec * u.deg)
        separations = center_coord.separation(sample_coords).deg
        assert np.all(separations <= 1.01)  # Small buffer for MOC approximation

    # We can create a sampler with no radii (just using the default for each table).
    moc_sampler2 = ApproximateMOCSampler.from_obstable_intersection(ops_tables, depth=10)
    assert moc_sampler2 is not None

    # We fail if the list is empty.
    with pytest.raises(ValueError):
        _ = ApproximateMOCSampler.from_obstable_intersection([], depth=10, radii=[], seed=45)

    # We fail if the list lengths do not match.
    with pytest.raises(ValueError):
        _ = ApproximateMOCSampler.from_obstable_intersection(ops_tables, depth=10, radii=[1.0], seed=45)



"""Wrappers for the models defined in EZTaoX.

https://github.com/LSST-AGN-Variability/EzTaoX
"""

import importlib

import numpy as np
from citation_compass import CiteClass

from lightcurvelynx.astro_utils.mag_flux import mag2flux
from lightcurvelynx.math_nodes.np_random import NumpyRandomFunc
from lightcurvelynx.models.physical_model import BandfluxModel


class EzTaoXWrapperModel(BandfluxModel, CiteClass):
    """A wrapper for an eztaox model.

    Parameterized values include:
      * dec - The object's declination in degrees. [from BasePhysicalModel]
      * distance - The object's luminosity distance in pc. [from BasePhysicalModel]
      * ra - The object's right ascension in degrees. [from BasePhysicalModel]
      * redshift - The object's redshift. [from BasePhysicalModel]
      * t0 - The t0 of the zero phase, date. [from BasePhysicalModel]
    Additional parameterized values are used for specific eztaox models.

    References
    ----------
    * Weixiang Yu et al., 2025 “Scalable and Robust Multiband Modeling of AGN Light Curves in Rubin-LSST"
      DOI: 10.48550/arXiv.2511.21479

    Attributes
    ----------
    kernel : eztaox kernel object
        An eztaox kernel object to use for the Gaussian process modeling of the light curve.
    num_kernel_params : int
        The number of kernel parameters for the eztaox kernel.
    num_filters : int
        The number of filters in the model.
    zero_mean : bool
        Whether to use a zero mean model.
    has_lag : bool
        Whether the model includes lag parameters.
    filter_idx : dict
        A mapping from filter name to an integer index.

    Parameters
    ----------
    kernel : eztaox kernel object
        An eztaox kernel object to use for the Gaussian process modeling of the light curve.
    baseline_mags : dict, optional
        A mapping from filter name to the setter baseline magnitude for that filter. If not
        provided, the model will use zero mean or the mean_func/mean_mag parameters.
        Default is None.
    log_kernel_param : list of setters, required
        Setters for each of the log kernel parameters. These must be in the order expected by
        the kernel functions.
    amp_scale_func : Callable, optional
        A callable amplitude scaling function, defaults to None.
    log_amp_scale : list of setters, optional
        Setters for the log amplitude scale for each filter (length N) if amp_scale_func is
        not provided.
        Default is None.
    zero_mean : bool
        Whether to use a zero mean model. If False then the program will try to use (in order):
        mean_func, mean (values), or a default mean function.
        Default is True.
    mean_func : Callable, optional
        If provided and zero_mean is False this is used to compute the mean function
        for bands.
        Default is None.
    mean_mag : list of setters, optional
        Setters for the mean magnitude for each filter except the first (length N-1) if
        zero_mean is False and mean_func is not provided.
        Default is None.
    has_lag : bool
        Whether the model includes lag parameters. If True then the program will try to
        use (in order): lag_func, lag (values), or a default lag function.
        Default is False.
    lag_func : Callable, optional
        If provided and has_lag is True this is used to compute the lag.
        Default is None.
    lag : list of setters, optional
        Setters for the lag for each filter except the first (length N) if has_lag is True
        and lag_func is not provided.
        Default is None.
    band_list : list, optional
        A list of band names in order. If not provided,
        the default list for ugrizy filters is used.
    seed_param : setter, optional
        A setter for the seed parameter to use for each run. If not provided, a random
        seed is generated for each run.
        Default is None.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    # Convenience mapping from filter name to index in the parameter list.
    _default_filter_idx = {"u": 0, "g": 1, "r": 2, "i": 3, "z": 4, "y": 5}

    def __init__(
        self,
        kernel,
        *,
        baseline_mags=None,
        log_kernel_param=None,
        amp_scale_func=None,
        log_amp_scale=None,
        zero_mean=True,
        mean_func=None,
        mean_mag=None,
        has_lag=False,
        lag_func=None,
        lag=None,
        band_list=None,
        seed_param=None,
        **kwargs,
    ):
        self._cached_data = {}
        super().__init__(**kwargs)

        # Confirm that the needed packages are installed.
        if importlib.util.find_spec("eztaox") is None:  # pragma: no cover
            raise ImportError(
                "The EzTaoX package is required to use the EzTaoXWrapperModel. "
                "Please install it from https://github.com/LSST-AGN-Variability/EzTaoX"
            )
        if importlib.util.find_spec("jax") is None:  # pragma: no cover
            raise ImportError(
                "JAX is required to use the EzTaoXWrapperModel class, please "
                "install with `pip install jax` or `conda install conda-forge::jax`"
            )

        # Store the kernel and filter index mapping.
        self.kernel = kernel
        if band_list is not None:
            self.filter_idx = {band: idx for idx, band in enumerate(band_list)}
        else:
            self.filter_idx = self._default_filter_idx
        self.num_filters = len(self.filter_idx)

        # Add the log kernel parameters to this model node. Since the order is
        # the important aspect, we just name them by index.
        if log_kernel_param is None:
            raise ValueError("The log_kernel_param parameter setters must be provided.")
        self.num_kernel_params = len(log_kernel_param)
        for i, setter in enumerate(log_kernel_param):
            self.add_parameter(f"eztaox_log_kernel_param_{i}", setter)

        # Add the amplitude scale information as either a function or parameters to this node.
        self._amp_scale_func = amp_scale_func
        if log_amp_scale is None and amp_scale_func is None:
            raise ValueError("One of either amp_scale_func or log_amp_scale must be provided.")
        self._has_log_amp_scale = log_amp_scale is not None
        if self._has_log_amp_scale:
            for i, setter in enumerate(log_amp_scale):
                self.add_parameter(f"eztaox_log_amp_scale_{i}", setter)
            if len(log_amp_scale) != self.num_filters:
                raise ValueError(
                    f"The number of log amplitude scale parameter setters {len(log_amp_scale)} "
                    f"must be equal to the number of filters {self.num_filters}."
                )

        # Store the mean magnitude parameters and callable if provided.
        self.zero_mean = zero_mean
        self._mean_func = mean_func  # The callable if it is provided.
        self._has_mean_vals = mean_mag is not None
        if self._has_mean_vals:
            if baseline_mags is not None:
                raise ValueError(
                    "If mean_mag parameter setters are provided, then baseline_mags cannot also be provided."
                )
            if len(mean_mag) != self.num_filters - 1:
                raise ValueError(
                    f"The number of mean parameter setters {len(mean_mag)} must be equal to the "
                    f"number of filters minus one {self.num_filters - 1}."
                )
            for i, setter in enumerate(mean_mag):
                self.add_parameter(f"eztaox_band_mean_{i}", setter)

        # Store the lag parameters if provided.
        self.has_lag = has_lag
        self._lag_func = lag_func  # The callable if it is provided.
        self._has_lag_values = lag is not None
        if self._has_lag_values:
            if len(lag) != self.num_filters:
                raise ValueError(
                    f"The number of lag parameter setters {len(lag)} must be equal to the "
                    f"number of filters {self.num_filters}."
                )
            for i, setter in enumerate(lag):
                self.add_parameter(f"eztaox_band_lag_{i}", setter)

        # Store the baseline magnitude information.
        for filter in self.filter_idx:
            param_name = f"eztaox_baseline_mag_{filter}"
            if baseline_mags is not None and filter in baseline_mags:
                self.add_parameter(param_name, baseline_mags[filter])
            else:
                self.add_parameter(param_name, 0.0)  # Default to 0.0 if not provided.

        # The seed used per run can be defined by the seed_param. If is not provided,
        # we randomly generate a seed for each run.
        if seed_param is None:
            seed_param = NumpyRandomFunc("integers", low=0, high=2**32 - 1)
        self.add_parameter("eztaox_seed_param", seed_param)

    def _compute_all_bandfluxes(self, times, filters, state):
        """Evaluate the model at the passband level for a single, given graph state and
        and all of the filters at once. We do this and cache the results to avoid
        recomputing the same model for each band.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filters : list of str
            The names of the filters (one at each time).
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
            This is not used in this model, but is required for the function signature.
        """
        if (
            "bandfluxes" in self._cached_data
            and state is self._cached_data.get("state")
            and np.array_equal(self._cached_data["filters"], filters)
            and np.array_equal(self._cached_data["times"], times)
        ):  # pragma: no cover
            return  # Nothing to do, the cache is valid.

        # Cache the input data (times and filters).
        self._cached_data = {}  # Clear the cache (just in case).
        self._cached_data["times"] = times
        self._cached_data["filters"] = filters
        self._cached_data["state"] = state

        # Import the dependencies that we need for this computation (these have all
        # been checked in the constructor).
        import jax
        import jax.numpy as jnp
        from eztaox.simulator import MultiVarSim

        # Shift the times to be relative to start at 0.
        delta_t = jnp.array(times) - jnp.min(jnp.array(times))  # Shift times to start at 0.

        # Extract the local parameters for this object from the full state object and build
        # the parameter dict needed by the simulator. This parameter dict must include:
        # - log_kernel_param a JAX numpy array of shape (num_kernel_params,)
        # - log_amp_scale a JAX numpy array of shape (num_filters,)
        # Optional it may also include:
        # - mean a JAX numpy array of shape (num_filters - 1,)
        # - lag a JAX numpy array of shape (num_filters,)
        local_params = self.get_local_params(state)
        init_params = {}
        init_params["log_kernel_param"] = jnp.array(
            [local_params[f"eztaox_log_kernel_param_{i}"] for i in range(self.num_kernel_params)]
        )
        if self._has_log_amp_scale:
            init_params["log_amp_scale"] = jnp.array(
                [local_params[f"eztaox_log_amp_scale_{i}"] for i in range(self.num_filters)]
            )
        if self._has_mean_vals:
            init_params["mean"] = jnp.array(
                [local_params[f"eztaox_band_mean_{i}"] for i in range(self.num_filters - 1)]
            )
        if self._has_lag_values:
            init_params["lag"] = jnp.array(
                [local_params[f"eztaox_band_lag_{i}"] for i in range(self.num_filters)]
            )

        # Create the simulator object using the given parameters and run it.
        sim = MultiVarSim(
            self.kernel,
            0.01,
            jnp.max(delta_t),  # The last time to simulate.
            self.num_filters,
            init_params=init_params,
            mean_func=self._mean_func,
            amp_scale_func=self._amp_scale_func,
            lag_func=self._lag_func,
            zero_mean=self.zero_mean,
            has_lag=self.has_lag,
        )

        # Compute the list of bands as integer indices for the simulator and save them to the cache.
        band_indices = jnp.array([self.filter_idx[f] for f in filters])

        # Compute the list of magnitudes for all given times, transform them to fluxes,
        # save them in the cache, and return them.
        _, mags = sim.fixed_input_fast(
            (delta_t, band_indices),  # Tuple of times and band indices
            jax.random.PRNGKey(local_params["eztaox_seed_param"]),  # Use the per-run seed.
        )

        # Add in the baseline magnitudes. Note we need to create a new copy of the array here,
        # because it is a JAX array and cannot be modified in place.
        mags = np.array(mags)
        for filter in self.filter_idx:
            baseline_mag = local_params[f"eztaox_baseline_mag_{filter}"]
            filter_mask = filters == filter
            mags[filter_mask] += baseline_mag

        # Convert to fluxes and cache.
        bandfluxes = mag2flux(mags)
        self._cached_data["bandfluxes"] = bandfluxes

    def compute_bandflux(self, times, filter, state):
        """Evaluate the model at the passband level for a single, given graph state and filter.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filter : str
            The name of the filter.
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
            This is not used in this model, but is required for the function signature.

        Returns
        -------
        bandflux : numpy.ndarray
            A length T array of band fluxes for this model in this filter.
        """
        if self._cached_data is None or state is not self._cached_data.get("state"):
            raise NotImplementedError(
                "The compute_bandflux method should not be called directly for the "
                "EzTaoXWrapperModel. Instead, use the evaluate_bandfluxes method which "
                "handles caching of the bandflux computations."
            )

        # Extract the bandfluxes from the entries with matching filters.
        filter_mask = self._cached_data["filters"] == filter
        return self._cached_data["bandfluxes"][filter_mask]

    def evaluate_bandfluxes(self, passband_or_group, times, filters, state, rng_info=None) -> np.ndarray:
        """Get the band fluxes for a given Passband or PassbandGroup.

        Parameters
        ----------
        passband_or_group : Passband or PassbandGroup or None
            The passband (or passband group) to use. Not used in this function.
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filters : numpy.ndarray or None
            A length T array of filter names. It may be None if
            passband_or_group is a Passband.
        state : GraphState or None
            An object mapping graph parameters to their values.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.

        Returns
        -------
        bandfluxes : numpy.ndarray
            A matrix of the band fluxes. If only one sample is provided in the GraphState,
            then returns a length T array. Otherwise returns a size S x T array where S is the
            number of samples in the graph state.
        """
        # Check if we need to sample the graph.
        if state is None:
            state = self.sample_parameters(num_samples=1, rng_info=rng_info)

        # Compute the bandfluxes using the cached method.
        self._compute_all_bandfluxes(times, filters, state)

        # Call the parent method to extract the bandfluxes in the right format.
        bandfluxes = super().evaluate_bandfluxes(passband_or_group, times, filters, state, rng_info=rng_info)

        # Clear the cache
        self._cached_data.clear()

        return bandfluxes



import astropy.units as u
import numpy as np
import pandas as pd
import pytest
from astropy.coordinates import SkyCoord
from cdshealpix import skycoord_to_healpix
from lightcurvelynx.obstable.argus_obstable import ArgusHealpixObsTable


def _ra_dec_to_healpix(ra, dec, nside=32):
    """Convert RA and Dec to HEALPix indices."""
    healpix_depth = int(np.log2(nside))
    coords = SkyCoord(ra=ra * u.deg, dec=dec * u.deg, frame="icrs")
    return skycoord_to_healpix(coords, healpix_depth)


def test_create_argus_obstable():
    """Create a minimal ArgusHealpixObsTable object and perform basic queries."""
    ra = np.array([15.0, 30.0, 15.0, 0.0, 15.0, 30.0])
    dec = np.array([10.0, -5.0, 10.0, 5.0, 10.0, -5.0])
    healpix = _ra_dec_to_healpix(ra, dec)
    values = {
        "epoch": 59000.0 + np.arange(len(ra)),
        "ra": ra,
        "dec": dec,
        "zp": np.ones_like(ra),
        "nside": np.full_like(ra, 32),
    }
    pdf = pd.DataFrame(values, index=healpix)

    # We fail if we are missing a healpix column or index.
    with pytest.raises(ValueError):
        _ = ArgusHealpixObsTable(pdf)

    pdf.index.name = "healpix"
    ops_data = ArgusHealpixObsTable(pdf)
    assert len(ops_data) == 6
    assert ops_data.healpix_nside == 32
    assert ops_data.healpix_depth == 5

    # We have all the attributes set at their default values.
    assert ops_data.survey_values["pixel_scale"] == 1.0
    assert ops_data.survey_values["radius"] == 52.0
    assert ops_data.survey_values["read_noise"] == 1.4
    assert ops_data.survey_values["survey_name"] == "Argus"

    # Check that we can extract the time bounds.
    t_min, t_max = ops_data.time_bounds()
    assert t_min == 59000.0
    assert t_max == 59005.0

    # We can query which columns the LSSTObsTable has by new or old name.
    assert "ra" in ops_data
    assert "dec" in ops_data
    assert "time" in ops_data
    assert "zp" in ops_data
    assert "healpix" in ops_data  # We move the index to a column
    assert "filter" in ops_data  # We add a default filter column

    # We can access columns directly as though it was a table.
    assert np.allclose(ops_data["ra"], values["ra"])
    assert np.allclose(ops_data["dec"], values["dec"])
    assert np.allclose(ops_data["time"], values["epoch"])

    # Check that we can determine which times a point is seen.
    assert np.array_equal(ops_data.range_search(15.0, 10.0), [0, 2, 4])
    assert np.array_equal(ops_data.range_search(30.0, -5.0), [1, 5])
    assert np.array_equal(ops_data.range_search(0.0, 5.0), [3])

    # Far away points are not seen.
    assert np.array_equal(ops_data.range_search(100.0, 80.0), [])
    assert np.array_equal(ops_data.range_search(15.0, -0.0), [])

    # Very close points fall in the same bucket, since they fall within an observed healpix pixel.
    assert np.array_equal(ops_data.range_search(15.00001, 9.99999), [0, 2, 4])
    assert np.array_equal(ops_data.range_search(29.999, -4.999), [1, 5])

    # We can filter the range search by time.
    assert np.array_equal(ops_data.range_search(15.0, 10.0, t_min=59001.0), [2, 4])
    assert np.array_equal(ops_data.range_search(15.0, 10.0, t_max=59001.0), [0])
    assert np.array_equal(ops_data.range_search(15.0, 10.0, t_min=59001.0, t_max=59003.0), [2])

    # We can do a vectorized range search.
    matches = ops_data.range_search(np.array([15.0, 30.0]), np.array([10.0, -5.0]))
    assert len(matches) == 2
    assert np.array_equal(matches[0], [0, 2, 4])
    assert np.array_equal(matches[1], [1, 5])

    # Check we can build a MOC.
    assert ops_data.build_moc() is not None


def test_create_argus_obstable_from_dict():
    """Create a ArgusHealpixObsTable object from a dictionary instead of a DataFrame."""
    ra = np.array([15.0, 30.0, 15.0, 0.0, 15.0, 30.0])
    dec = np.array([10.0, -5.0, 10.0, 5.0, 10.0, -5.0])
    healpix = _ra_dec_to_healpix(ra, dec)
    values = {
        "epoch": 59000.0 + np.arange(len(ra)),
        "ra": ra,
        "dec": dec,
        "zp": np.ones_like(ra),
        "healpix": healpix,
        "nside": np.full_like(ra, 32),
    }

    table = ArgusHealpixObsTable(values)
    assert len(table) == 6


def test_create_argus_obstable_alternate():
    """Create a ArgusHealpixObsTable object with healpix as a column instead of index."""
    ra = np.array([15.0, 30.0, 15.0, 0.0, 15.0, 30.0])
    dec = np.array([10.0, -5.0, 10.0, 5.0, 10.0, -5.0])
    healpix = _ra_dec_to_healpix(ra, dec)
    values = {
        "epoch": 59000.0 + np.arange(len(ra)),
        "ra": ra,
        "dec": dec,
        "zp": np.ones_like(ra),
        "healpix": healpix,  # We move the index to a column
    }
    pdf = pd.DataFrame(values)

    # We fail without nside information.
    with pytest.raises(ValueError):
        _ = ArgusHealpixObsTable(pdf)

    table = ArgusHealpixObsTable(pdf, nside=32)
    assert len(table) == 6

    # We fail if we have inconsistent nside information.
    pdf["nside"] = np.array([32, 32, 16, 32, 32, 32])
    with pytest.raises(ValueError):
        _ = ArgusHealpixObsTable(pdf)


def test_argus_obstable_footprint():
    """We cannot use a footprint for the argus obstable since it is already healpix."""
    ra = np.array([15.0, 30.0, 15.0, 0.0, 15.0, 30.0])
    dec = np.array([10.0, -5.0, 10.0, 5.0, 10.0, -5.0])
    healpix = _ra_dec_to_healpix(ra, dec)
    values = {
        "epoch": 59000.0 + np.arange(len(ra)),
        "ra": ra,
        "dec": dec,
        "zp": np.ones_like(ra),
        "healpix": healpix,  # We move the index to a column
    }
    pdf = pd.DataFrame(values)
    table = ArgusHealpixObsTable(pdf, nside=32)
    assert not table.uses_footprint()

    # We fail if we even call set_detector_footprint.
    with pytest.raises(NotImplementedError):
        table.set_detector_footprint(None)


def test_argus_obstable_noise():
    """Test that we can automatically compute the zero point for a ArgusHealpixObsTable
    and use that to compute the bandflux_error
    """
    ra = np.array([15.0, 30.0, 15.0, 0.0, 15.0, 30.0])
    dec = np.array([10.0, -5.0, 10.0, 5.0, 10.0, -5.0])
    healpix = _ra_dec_to_healpix(ra, dec)
    values = {
        "epoch": 59000.0 + np.arange(len(ra)),
        "ra": ra,
        "dec": dec,
        "healpix": healpix,
        # To derive zeropoint we need dark_electrons, exptime, seeing
        "dark_electrons": np.full_like(ra, 0.405696),
        "exptime": np.full_like(ra, 60.0),
        "limmag": np.full_like(ra, 20.569349),
        "seeing": np.full_like(ra, 0.804956),
        "sky_electrons": np.full_like(ra, 243.357877),
    }
    pdf = pd.DataFrame(values)
    table = ArgusHealpixObsTable(pdf, nside=32)

    assert "zp" in table
    assert np.all(table["zp"] > 15.0)

    # Check that we can compute the bandflux error for a given bandflux.
    bandfluxes = np.array([1000.0, 500.0, 200.0])
    index = np.array([0, 2, 4])
    bandflux_err = table.bandflux_error_point_source(bandflux=bandfluxes, index=index)
    assert len(bandflux_err) == len(bandfluxes)
    assert np.all(bandflux_err > 0.0)



"""Wrappers for the models defined in pyLIMA.

https://github.com/ebachelet/pyLIMA
"""

import warnings

import numpy as np
from citation_compass import CiteClass

from lightcurvelynx.astro_utils.mag_flux import Mag2FluxNode
from lightcurvelynx.consts import MJD_TO_JD_OFFSET
from lightcurvelynx.models.physical_model import BandfluxModel
from lightcurvelynx.utils.io_utils import SquashOutput


def _load_pylima_model_class(model_name):
    """Load a pyLIMA model class by name.

    Parameters
    ----------
    model_name : str
        The name of the pyLIMA model type to load. E.g., 'PSPL'

    Returns
    -------
    model_class : class
        The pylima model class.
    """
    try:
        from pyLIMA import models
    except ImportError as err:  # pragma: no cover
        raise ImportError(
            "The pyLIMA package is required to use the PyLIMAWrapperModel. Please install it. "
            "See https://pylima.readthedocs.io/en/latest/ for instructions."
        ) from err

    # The package and the class uses different naming conventions with the
    # package having '_model' suffix and the class having 'model' suffix.
    model_package = getattr(models, f"{model_name}_model")
    model_class = getattr(model_package, f"{model_name}model")
    return model_class


class PyLIMAWrapperModel(BandfluxModel, CiteClass):
    """A wrapper for single pyLIMA models (one model type).

    Parameterized values include:
      * dec - The object's declination in degrees. [from BasePhysicalModel]
      * distance - The object's luminosity distance in pc. [from BasePhysicalModel]
      * ra - The object's right ascension in degrees. [from BasePhysicalModel]
      * redshift - The object's redshift. [from BasePhysicalModel]
      * t0 - The t0 of the zero phase, date. [from BasePhysicalModel]
    Additional parameterized values are used for specific PyLIMA models.

    Parameters
    ----------
    model_info : str or class
        The name of the pyLIMA model class to use in the simulation or the class itself.
    source_mags : dict
        A mapping from filter names to source magnitudes for the microlensed source.
    blend_mags : dict, optional
        A mapping from filter names to blending magnitudes for the microlensed source.
    pylima_model_params : dict, optional
        A dictionary of additional pyLIMA parameters for the model, such as
        'u0', 'tE', 'rho', etc. If a parameter is already added to the model,
        its value will be updated instead.
    parallax_model : str, optional
        The pyLIMA parallax model type: 'None', 'Annual', 'Terrestrial', or 'Full'.
        The times for the parallax are automatically set during the evaluation.
    blend_flux_parameter : str, optional
        The pyLIMA blend flux parameter type. Currently only 'fblend' is supported.
        See also https://github.com/lincc-frameworks/lightcurvelynx/issues/691
    time_frame_offset : float, optional
        PyLIMA models use JD for time while users may specify any time system. This offset
        is added to the input times to convert them to JD. By default, this is set to
        MJD_TO_JD_OFFSET to convert from MJD to JD.
    observer_location : str, optional
        The location of the observer. Default is 'Earth'.
    **kwargs : dict, optional
        Any additional keyword arguments.

    Attributes
    ----------
    filters : list
        The list of filters supported by this model.
    parallax_model : str, optional
        The pyLIMA parallax model type: 'None', 'Annual', 'Terrestrial', or 'Full'.
    blend_flux_parameter : str, optional
        The pyLIMA blend flux parameter type. Currently only 'fblend' is supported.
        See also https://github.com/lincc-frameworks/lightcurvelynx/issues/691
    time_frame_offset: float, optional
        PyLIMA models use JD for time while users may specify any time system. This offset
        is added to the input times to convert them to JD.
    observer_location : str, optional
        The location of the observer.
    """

    def __init__(
        self,
        model_info,
        source_mags,
        *,
        blend_mags=None,
        pylima_params=None,
        parallax_model="None",
        blend_flux_parameter="fblend",
        time_frame_offset=MJD_TO_JD_OFFSET,
        observer_location="Earth",
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.time_frame_offset = time_frame_offset
        self.observer_location = observer_location

        # Save the pyLIMA parameters and add the corresponding model parameters.
        self.parallax_model = parallax_model
        self.blend_flux_parameter = blend_flux_parameter
        if blend_flux_parameter != "fblend":
            raise ValueError(
                f"Invalid blend_flux_parameter '{blend_flux_parameter}'. "
                f"Currently only 'fblend' is supported. See: "
                "https://github.com/lincc-frameworks/lightcurvelynx/issues/691"
            )

        # Add each source flux as a parameter by converting the input magnitudes.
        self.filters = list(source_mags.keys())
        for filter_name, mag in source_mags.items():
            self.add_parameter(
                f"fsource_{filter_name}",
                Mag2FluxNode(mag),
            )

        # Do the same for the (optional) blending magnitudes.
        if blend_mags is None:
            blend_mags = {}
        for filter_name in self.filters:
            param_name = f"{blend_flux_parameter}_{filter_name}"
            param_val = Mag2FluxNode(blend_mags[filter_name]) if filter_name in blend_mags else 0.0
            self.add_parameter(param_name, param_val)

        # Add any of the pyLIMA parameters from the pylima_params dictionary.
        if pylima_params is not None:
            for name, value in pylima_params.items():
                if name not in self.setters:
                    self.add_parameter(name, value)
                else:  # pragma: no cover
                    warnings.warn(f"Parameter {name} already exists in the model. Overriding its value.")
                    self.set_parameter(name, value)

        # Save the model class and create a model object. We allow the user
        # to pass in a class to simplify testing when PyLIMA is not installed.
        if isinstance(model_info, str):
            self._model_class = _load_pylima_model_class(model_info)
        else:
            self._model_class = model_info

        # Create a dummy pyLIMA model instance to get the parameter names and
        # check that they are all added. We use an arbitrary test time for parallax.
        test_time = 60676.0 + self.time_frame_offset
        with SquashOutput():
            event = self.make_pylima_event(
                ra=0.0,
                dec=0.0,
                filter="r",
                times=np.array([test_time]),
            )
            model = self._model_class(
                event,
                parallax=[self.parallax_model, test_time],
                blend_flux_parameter=self.blend_flux_parameter,
            )

        expected_params_map = model.pyLIMA_standards_dictionnary
        for name in expected_params_map:
            if name not in self.setters:
                if name in kwargs:
                    self.add_parameter(name, kwargs[name])
                else:
                    raise ValueError(
                        f"The pyLIMA model '{self._model_class.__name__}' uses parameter {name} "
                        f"but this was not provided as an argument or added as a model parameter."
                    )

    def make_pylima_event(self, ra, dec, filter=None, times=None):
        """Create a pyLIMA event object and attach a telescope if filter and times are given.

        Parameters
        ----------
        ra : float
            The right ascension of the event in degrees.
        dec : float
            The declination of the event in degrees.
        filter : str, optional
            The name of the filter for the telescope to attach.
        times : numpy.ndarray, optional
            A length T array of observer frame timestamps in JD for the telescope to attach.
        """
        try:
            from pyLIMA import event
            from pyLIMA.simulations import simulator
        except ImportError as err:  # pragma: no cover
            raise ImportError(
                "The pyLIMA package is required to use the PyLIMAWrapperModel. Please install it. "
                "See https://pylima.readthedocs.io/en/latest/ for instructions."
            ) from err

        pylima_event = event.Event(ra=ra, dec=dec)
        pylima_event.name = "LightCurveLynx_Event"

        if filter is not None and times is not None:
            # Create a telescope object for the given filter.
            tel = simulator.simulate_a_telescope(
                name=filter,
                location=self.observer_location,
                timestamps=times,
                astrometry=False,
            )
            pylima_event.telescopes.append(tel)

        return pylima_event

    def compute_bandflux(self, times, filter, state):
        """Evaluate the model at the passband level for a single, given graph state and filter.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filter : str
            The name of the filter.
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
            This is not used in this model, but is required for the function signature.

        Returns
        -------
        bandflux : numpy.ndarray
            A length T array of band fluxes for this model in this filter.
        """
        # Check that the filter is supported.
        if filter not in self.filters:
            raise ValueError(f"Filter '{filter}' is not supported by this model (mags / blend).")

        params = self.get_local_params(state)

        try:
            from pyLIMA.simulations import simulator
        except ImportError as err:  # pragma: no cover
            raise ImportError(
                "The pyLIMA package is required to use the PyLIMAWrapperModel. Please install it. "
                "See https://pylima.readthedocs.io/en/latest/ for instructions."
            ) from err

        # Compute the T0 in JD (which PyLIMA uses) from the MJD value in our parameters.
        t0_jd = params["t0"] + self.time_frame_offset
        times_jd = times + self.time_frame_offset

        # Squash pyLIMA's print output.
        with SquashOutput():
            # Create the PyLIMA event object with the attached telescope and a model instance.
            current_event = self.make_pylima_event(
                params["ra"],
                params["dec"],
                filter=filter,
                times=times_jd,
            )

            # Create the model instance.
            model = self._model_class(
                current_event,
                parallax=[self.parallax_model, t0_jd],
                blend_flux_parameter=self.blend_flux_parameter,
            )

            # Get the expected parameter mapping and create the ordered parameter list.
            expected_params_map = model.pyLIMA_standards_dictionnary
            ordered_values = [0.0] * len(expected_params_map)
            for name, index in expected_params_map.items():
                if name == "t0":
                    # We need to special case t0 because LightCurveLynx uses MJD and PyLIMA uses JD.
                    ordered_values[index] = t0_jd
                elif name in params:
                    ordered_values[index] = params[name]
            pyLIMA_params = model.compute_pyLIMA_parameters(ordered_values)

            # Simulate the lightcurve without noise.
            simulator.simulate_lightcurve(model, pyLIMA_params, add_noise=False)
            fluxes = current_event.telescopes[0].lightcurve["flux"]

        return fluxes



"""Samplers used for generating (RA, dec) coordinates."""

import logging
from pathlib import Path

import numpy as np
from astropy.coordinates import Angle, SkyCoord
from cdshealpix.nested import healpix_to_skycoord
from citation_compass import CiteClass
from mocpy import MOC

from lightcurvelynx.math_nodes.given_sampler import TableSampler
from lightcurvelynx.math_nodes.np_random import NumpyRandomFunc
from lightcurvelynx.obstable.obs_table import ObsTable
from lightcurvelynx.utils.plotting import plot_moc


class UniformRADEC(NumpyRandomFunc):
    """A FunctionNode that uniformly samples (RA, dec) over a sphere,

    Attributes
    ----------
    use_degrees : bool
        The default return unit. If True returns samples in degrees.
        Otherwise, if False, returns samples in radians.
    """

    def __init__(self, outputs=None, seed=None, use_degrees=True, **kwargs):
        self.use_degrees = use_degrees

        # Override key arguments. We create a uniform sampler function, but
        # won't need it because the subclass overloads compute().
        func_name = "uniform"
        outputs = ["ra", "dec"]
        super().__init__(func_name, outputs=outputs, seed=seed, **kwargs)

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Return the given values.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        results : any
            The result of the computation. This return value is provided so that testing
            functions can easily access the results.
        """
        rng = rng_info if rng_info is not None else self._rng

        # Generate the random (RA, dec) lists.
        ra = rng.uniform(0.0, 2.0 * np.pi, size=graph_state.num_samples)
        dec = np.arcsin(rng.uniform(-1.0, 1.0, size=graph_state.num_samples))
        if self.use_degrees:
            ra = np.degrees(ra)
            dec = np.degrees(dec)

        # If we are generating a single sample, return floats.
        if graph_state.num_samples == 1:
            ra = ra[0]
            dec = dec[0]

        # Set the outputs and return the results. This takes the place of
        # function node's _save_results() function because we know the outputs.
        graph_state.set(self.node_string, "ra", ra)
        graph_state.set(self.node_string, "dec", dec)
        return [ra, dec]


class ObsTableRADECSampler(TableSampler):
    """A FunctionNode that randomly samples RA and dec (and any extra columns) from
    around given data. This node is used for both sampling from a list of pointings
    (where the radius is the field of view) or sampling from a list of true objects.
    This sampling strategy will return a visit-weighted distribution, so areas of the
    sky that are observed more often will be sampled more often.

    Parameters
    ----------
    data : ObsTable, Pandas DataFrame, NestedFrame, or dict
        The data to use for sampling. Must contain 'ra' and 'dec' columns.
    extra_cols : list of str, optional
        A list of extra column names to include in the sampling.
        Default: None
    radius : float, optional
        The sampling radius around the given points. If the points represent the
        center of a pointing, this is the radius of the field of view in degrees.
        If the points represent exact true objects, this can be a noise factor.
        Use 0.0 to return the exact given points.
        If None and data is an ObsTable, uses the value from the ObsTable.
        Default: None
    **kwargs : dict, optional
        Additional keyword arguments to pass to the parent class constructor.
    """

    def __init__(self, data, *, extra_cols=None, radius=None, **kwargs):
        if isinstance(data, ObsTable):
            if radius is None:
                radius = data.radius
            else:
                logging.getLogger(__name__).info(
                    f"Using provided radius {radius} instead of ObsTable radius {data.radius}."
                )
        if radius is None or radius < 0.0:
            raise ValueError(f"Invalid radius: {radius}")
        self.radius = radius

        # Start with RA, dec, and (optionally) time.
        data_dict = {
            "ra": data["ra"],
            "dec": data["dec"],
        }
        if "time" in data:
            data_dict["time"] = data["time"]

        # Add any extra columns (without duplicates).
        if extra_cols is not None:
            for col in extra_cols:
                if col not in data_dict:
                    data_dict[col] = data[col]

        super().__init__(data_dict, in_order=False, **kwargs)

    @classmethod
    def from_hats(cls, path, *, radius=None, extra_cols=None, **kwargs):
        """Create a GivenRADECSampler from the observations in a HATS Catalog.

        Note
        ----
        If you have an existing Dask client, it may be used.
        See LSDB documentation for details: https://docs.lsdb.io/en/latest/

        Parameters
        ----------
        path : str or Path
            The base path of the HATS data directory.
        radius : float, optional
            The sampling radius (noise factor) in degrees. Use 0.0 to return
            the exact pointings.
            Default: 0.0
        extra_cols : list of str, optional
            A list of extra column names to include in the sampling.
            Default: None
        **kwargs : dict, optional
            Additional keyword arguments to pass to the constructor.

        Returns
        -------
        GivenRADECSampler
            The created GivenRADECSampler object.
        """
        # See if the (optional) LSDB package is installed.
        try:
            from lsdb import read_hats
        except ImportError as err:  # pragma: no cover
            raise ImportError(
                "The lsdb package is required to read HATS catalogs. "
                "Please install it via 'pip install lsdb' or 'conda install conda-forge::lsdb'."
            ) from err

        # Compute the full list of columns to load from HATS.
        cols_to_load = ["ra", "dec"]
        if extra_cols is not None:
            cols_to_load.extend(extra_cols)
        columns = list(set(cols_to_load))  # Remove any duplicates.

        data = read_hats(path, columns=columns).compute()
        return cls(data, extra_cols=extra_cols, radius=radius, **kwargs)

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Return the given values.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        results : any
            The result of the computation. This return value is provided so that testing
            functions can easily access the results.
        """
        # Sample the center RA, dec, and times without the radius. Results is a vector
        # of arrays, one per output column. The first two are guaranteed to be RA and dec.
        results = super().compute(graph_state, rng_info=rng_info, **kwargs)

        if self.radius > 0.0:
            rng = rng_info if rng_info is not None else np.random.default_rng()
            center = SkyCoord(ra=results[0], dec=results[1], unit="deg")

            # Add an offset from the center of the pointing defined by an offset angle (phi)
            # and offset radius (theta).  Both are defined in radians.
            offset_dir = rng.uniform(0.0, 2.0 * np.pi, size=graph_state.num_samples)
            cos_radius = np.cos(np.radians(self.radius))
            offset_amt = np.arccos(rng.uniform(cos_radius, 1.0, size=graph_state.num_samples))
            new_coords = center.directional_offset_by(
                position_angle=Angle(offset_dir, "radian"),
                separation=Angle(offset_amt, "radian"),
            )

            # Replace the results' RA and dec with the new pointing (in degrees).
            results[0] = new_coords.ra.deg
            results[1] = new_coords.dec.deg

            # Resave the results to transfer them to the graph state.
            self._save_results(results, graph_state)

        return results


class ObsTableUniformRADECSampler(NumpyRandomFunc):
    """A FunctionNode that samples RA and dec uniformly from the area covered
    by an ObsTable.  RA and dec are returned in degrees.

    Note
    ----
    This uses rejection sampling where it randomly guesses an (RA, dec) then checks if that
    point falls within the survey. If not, it repeats the process until it finds a valid point
    or reaches `max_iterations` iterations (then returns the last sample). This sampling method
    can be quite slow or even generate out-of-survey samples if the coverage is small.

    Attributes
    ----------
    data : ObsTable
        The ObsTable object to use for sampling.
    radius : float
        The radius of the field of view of the observations in degrees.
    max_iterations : int
        The maximum number of iterations to perform. Default: 1000

    Parameters
    ----------
    data : ObsTable
        The ObsTable object to use for sampling.
    radius : float, optional
        The search radius around the center of the pointing. If None, uses the
        value from the ObsTable.
    outputs : list of str, optional
        The list of output names. Default: ["ra", "dec"]
    seed : int, optional
        The random seed to use for the internal random number generator. Default: None
    max_iterations : int, optional
        The maximum number of iterations to perform. Default: 1000
    **kwargs : dict, optional
        Additional keyword arguments to pass to the parent class constructor.
    """

    def __init__(self, data, *, radius=None, outputs=None, seed=None, max_iterations=1000, **kwargs):
        if isinstance(data, ObsTable):
            if radius is None:
                radius = data.radius
            else:
                logging.getLogger(__name__).info(
                    f"Using provided radius {radius} instead of ObsTable radius {data.radius}."
                )
        if radius is None or radius < 0.0:
            raise ValueError(f"Invalid radius: {radius}")
        self.radius = radius

        if len(data) == 0:  # pragma: no cover
            raise ValueError("ObsTable data cannot be empty.")
        self.data = data

        if max_iterations <= 0:
            raise ValueError("Invalid max_iterations: {max_iterations}")
        self.max_iterations = max_iterations

        # Override key arguments. We create a uniform sampler function, but
        # won't need it because the subclass overloads compute().
        func_name = "uniform"
        outputs = ["ra", "dec"]
        super().__init__(func_name, outputs=outputs, seed=seed, **kwargs)

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Return the given values.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        (ra, dec) : tuple of floats or np.ndarray
            If a single sample is generated, returns a tuple of floats. Otherwise,
            returns a tuple of np.ndarrays.
        """
        rng = rng_info if rng_info is not None else self._rng

        ra = np.zeros(graph_state.num_samples)
        dec = np.zeros(graph_state.num_samples)
        mask = np.full(graph_state.num_samples, False)
        num_missing = graph_state.num_samples

        # Rejection sampling to ensure the samples are within the ObsTable coverage.
        # This can take many iterations if the coverage is small.
        iter_num = 1
        while num_missing > 0 and iter_num <= self.max_iterations:
            # Generate new samples for the missing ones.
            new_ra = np.degrees(rng.uniform(0.0, 2.0 * np.pi, size=num_missing))
            new_dec = np.degrees(np.arcsin(rng.uniform(-1.0, 1.0, size=num_missing)))
            ra[~mask] = new_ra
            dec[~mask] = new_dec

            # Check if the new samples are within the ObsTable coverage. We don't
            # need to recheck the ones that were already valid.
            mask[~mask] = self.data.is_observed(new_ra, new_dec, radius=self.radius)
            num_missing = np.sum(~mask)
            iter_num += 1

        # If we are generating a single sample, return floats.
        if graph_state.num_samples == 1:
            ra = ra[0]
            dec = dec[0]

        # Set the outputs and return the results. This takes the place of
        # function node's _save_results() function because we know the outputs.
        graph_state.set(self.node_string, "ra", ra)
        graph_state.set(self.node_string, "dec", dec)
        return (ra, dec)


class ApproximateMOCSampler(NumpyRandomFunc, CiteClass):
    """A FunctionNode that samples RA and dec (approximately) from the coverage of
    a MOCPy Multi-Order Coverage Map object.

    The depth parameter controls the approximation level. Higher depths provide
    better accuracy but require more memory and computation time. We recommend at
    least depth=12 for reasonable accuracy.

    References
    ----------
    * MOCPY: https://github.com/cds-astro/mocpy/
    * CDS Healpix: https://github.com/cds-astro/cds-healpix-python
    * MOC: Pierre Fernique, Thomas Boch, Tom Donaldson, Daniel Durand , Wil O'Mullane, Martin Reinecke,
    and Mark Taylor. MOC - HEALPix Multi-Order Coverage map Version 1.0. IVOA Recommendation 02 June 2014,
    pages 602, Jun 2014. doi:10.5479/ADS/bib/2014ivoa.spec.0602F.

    Attributes
    ----------
    healpix_list : list of int
        The list of healpix pixel IDs that cover the MOC at the given depth.
    depth : int
        The healpix depth to use as an approximation. Must be [2, 29].
    """

    def __init__(self, moc, *, outputs=None, seed=None, depth=12, **kwargs):
        if depth < 2 or depth > 29:
            raise ValueError(f"Depth must be [2, 29]. Received {depth}")
        self.depth = depth
        self.healpix_list = moc.to_order(depth).flatten()

        # Override key arguments. We create a uniform sampler function, but
        # won't need it because the subclass overloads compute().
        func_name = "uniform"
        outputs = ["ra", "dec"]
        super().__init__(func_name, outputs=outputs, seed=seed, **kwargs)

    @classmethod
    def from_file(cls, filename, format="fits", **kwargs):
        """Create an ApproximateMOCSampler from a MOC file.

        This file can be created from a mocpy.MOC object using its save() function.

        Parameters
        ----------
        filename : str or Path
            The path to the MOC file. Supported formats include FITS, JSON, and ASCII.
        format : str, optional
            The format of the MOC file. Supported formats include 'fits', 'json', and
            'ascii'. Default is 'fits'.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the constructor.

        Returns
        -------
        ApproximateMOCSampler
            The created ApproximateMOCSampler object.
        """
        filename = Path(filename)
        if not filename.is_file():
            raise FileNotFoundError(f"MOC file not found: {filename}")

        moc = MOC.load(filename, format=format)
        return cls(moc, **kwargs)

    @classmethod
    def from_obstable(
        cls,
        obstable,
        *,
        depth=12,
        use_footprint=False,
        radius=None,
        **kwargs,
    ):
        """Create an ApproximateMOCSampler from an ObsTable object.

        The depth parameter controls the approximation level. Higher depths provide
        better accuracy but require more memory and computation time. We recommend at
        least depth=12 for reasonable accuracy.

        Parameters
        ----------
        obstable : ObsTable
            The ObsTable object to use for creating the MOC.
        depth : int, optional
            The healpix depth to use as an approximation. Must be [2, 29].
            Default: 12
        radius : float, optional
            The radius to use for each image (in degrees). Only used if use_footprint
            is False. If None, the radius from the survey values will be used.
        use_footprint : bool, optional
            Whether to use the detector footprint to build the MOC. If True, the
            footprint will be used to compute the MOC regions for each pointing.
            If False, a simple cone with the given radius will be used.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the constructor.

        Returns
        -------
        ApproximateMOCSampler
            The created ApproximateMOCSampler object.
        """
        moc = obstable.build_moc(
            radius=radius,
            use_footprint=use_footprint,
            duplicate_threshold=10.0 / 3600.0,  # 10 arcsec
            max_depth=depth,
        )
        return cls(moc, depth=depth, **kwargs)

    @classmethod
    def from_obstable_union(
        cls,
        obstables,
        *,
        depth=12,
        use_footprint=False,
        radii=None,
        **kwargs,
    ):
        """Create an ApproximateMOCSampler from the union of the footprints in an ObsTable.

        The depth parameter controls the approximation level. Higher depths provide
        better accuracy but require more memory and computation time. We recommend at
        least depth=12 for reasonable accuracy.

        Parameters
        ----------
        obstables : list of ObsTable
            The list of ObsTable objects to use for creating the MOC. The union of
            their coverages will be used.
        depth : int, optional
            The healpix depth to use as an approximation. Must be [2, 29].
            Default: 12
        radii : list of float, optional
            The radius to use for each image (in degrees). Only used if use_footprint
            is False. If None, the radius from the survey values will be used.
        use_footprint : bool, optional
            Whether to use the detector footprint to build the MOC. If True, the
            footprint will be used to compute the MOC regions for each pointing.
            If False, a simple cone with the given radius will be used.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the constructor.

        Returns
        -------
        ApproximateMOCSampler
            The created ApproximateMOCSampler object.
        """
        if len(obstables) == 0:
            raise ValueError("At least one ObsTable must be provided.")
        if radii is None:
            radii = [None] * len(obstables)
        elif len(radii) != len(obstables):
            raise ValueError("Length of radii must match length of obstables.")

        moc = obstables[0].build_moc(
            radius=radii[0],
            use_footprint=use_footprint,
            duplicate_threshold=10.0 / 3600.0,  # 10 arcsec
            max_depth=depth,
        )
        for i in range(1, len(obstables)):
            next_moc = obstables[i].build_moc(
                radius=radii[i],
                use_footprint=use_footprint,
                duplicate_threshold=10.0 / 3600.0,  # 10 arcsec
                max_depth=depth,
            )
            moc = moc.union(next_moc)

        return cls(moc, depth=depth, **kwargs)

    @classmethod
    def from_obstable_intersection(
        cls,
        obstables,
        *,
        depth=12,
        use_footprint=False,
        radii=None,
        **kwargs,
    ):
        """Create an ApproximateMOCSampler from the intersection of the footprints in an ObsTable.

        The depth parameter controls the approximation level. Higher depths provide
        better accuracy but require more memory and computation time. We recommend at
        least depth=12 for reasonable accuracy.

        Parameters
        ----------
        obstables : list of ObsTable
            The list of ObsTable objects to use for creating the MOC. The intersection of
            their coverages will be used.
        depth : int, optional
            The healpix depth to use as an approximation. Must be [2, 29].
            Default: 12
        radii : list of float, optional
            The radius to use for each image (in degrees). Only used if use_footprint
            is False. If None, the radius from the survey values will be used.
        use_footprint : bool, optional
            Whether to use the detector footprint to build the MOC. If True, the
            footprint will be used to compute the MOC regions for each pointing.
            If False, a simple cone with the given radius will be used.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the constructor.

        Returns
        -------
        ApproximateMOCSampler
            The created ApproximateMOCSampler object.
        """
        if len(obstables) == 0:
            raise ValueError("At least one ObsTable must be provided.")
        if radii is None:
            radii = [None] * len(obstables)
        elif len(radii) != len(obstables):
            raise ValueError("Length of radii must match length of obstables.")

        moc = obstables[0].build_moc(
            radius=radii[0],
            use_footprint=use_footprint,
            duplicate_threshold=10.0 / 3600.0,  # 10 arcsec
            max_depth=depth,
        )
        for i in range(1, len(obstables)):
            next_moc = obstables[i].build_moc(
                radius=radii[i],
                use_footprint=use_footprint,
                duplicate_threshold=10.0 / 3600.0,  # 10 arcsec
                max_depth=depth,
            )
            moc = moc.intersection(next_moc)

        return cls(moc, depth=depth, **kwargs)

    def plot_footprint(
        self,
        *,
        depth=None,
        fig=None,
        ax=None,
        **kwargs,
    ):
        """Plot the MOC footprint using matplotlib.

        Parameters
        ----------
        depth : int, optional
            The healpix depth to use for plotting. If None, uses the depth of the sampler.
        fig : matplotlib.figure.Figure, optional
            An existing matplotlib figure to use. If None, a new figure is created.
        ax : matplotlib.pyplot.Axes or None, optional
            The axes to use for the plot. If None, new axes will be created on the figure.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the plot_moc function.

        Returns
        -------
        fig: matplotlib.figure.Figure
            The figure containing the plot.
        ax: matplotlib.pyplot.Axes
            The axes containing the plot.
        """
        if depth is None:
            depth = self.depth

        # Build a new MOC from the healpix list at the given depth.
        moc = MOC.from_healpix_cells(self.healpix_list, depth=self.depth, max_depth=depth)

        fig, ax = plot_moc(moc, fig=fig, ax=ax, **kwargs)
        return fig, ax

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Return the given values.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        (ra, dec) : tuple of floats or np.ndarray
            If a single sample is generated, returns a tuple of floats. Otherwise,
            returns a tuple of np.ndarrays.
        """
        rng = rng_info if rng_info is not None else self._rng

        # Choose a starting pixel ID for each sample. Then randomly traverse
        # down the healpix tree by moving to one of the children pixels until
        # we reach level=29 (approximately 4.5 * 10^18 possible locations).
        pixel_ids = rng.choice(self.healpix_list, size=graph_state.num_samples).astype(np.uint64)
        start_pixel_ids29 = np.left_shift(pixel_ids, 2 * (29 - self.depth))
        offset_range = np.uint64(1) << np.uint64(2 * (29 - self.depth))
        pixel_ids29 = start_pixel_ids29 + rng.integers(
            offset_range, size=graph_state.num_samples, dtype=np.uint64
        )

        # Convert back the healpix centers to RA and dec.
        coords = healpix_to_skycoord(pixel_ids29, depth=29)
        ra = coords.ra.deg
        dec = coords.dec.deg

        # If we are generating a single sample, return floats.
        if graph_state.num_samples == 1:
            ra = ra[0]
            dec = dec[0]

        # Set the outputs and return the results. This takes the place of
        # function node's _save_results() function because we know the outputs.
        graph_state.set(self.node_string, "ra", ra)
        graph_state.set(self.node_string, "dec", dec)
        return (ra, dec)



"""Functions for extrapolating flux past the end of a model's range of valid
phases or wavelengths using flux = f(time, wavelengths).
"""

import abc

import numpy as np

from lightcurvelynx.astro_utils.mag_flux import flux2mag, mag2flux


class FluxExtrapolationModel(abc.ABC):
    """The base class for the flux extrapolation methods.

    Attributes
    ----------
    nfit : int
        The number of points to be used for extrapolation. (Default is 1)
    """

    @abc.abstractmethod
    def __init__(self):
        # By default we compute extrapolation using the last valid point. This can be changed
        # by setting nfit to a larger number (e.g. in LinearFit).
        self.nfit = 1

    @abc.abstractmethod
    def _extrapolate(self, last_values, last_fluxes, query_values):
        """Evaluate the extrapolation given the last valid points(s) and a list of new
        query points.

        Note
        ----
        This function does not care which axis is being extrapolated. The returned values are
        always len(query_values) x len(last_fluxes) and may need to be transposed by the calling
        function.

        Parameters
        ----------
        last_values : float or np.ndarray
            The last valid value or a length L array of the last valid values along the extrapolation
            axis at which the fluxes were predicted
            (e.g., wavelength in AA or time in days).
        last_fluxes : numpy.ndarray
            A length N x L array of the flux values at the last valid time or wavelength (in nJy).
        query_values : numpy.ndarray
            A length M array of values along the extrapolation axis (times in days or wavelengths
            in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A N x M matrix of extrapolated values. Where M is the number of query points and
            N is the number of flux values at the last valid point.
        """
        raise NotImplementedError("Subclasses must implement this method.")  # pragma: no cover

    def extrapolate_time(self, last_times, last_fluxes, query_times):
        """Extrapolate along the time axis.

        Parameters
        ----------
        last_times : float or np.ndarray
            A length T1 array of the last valid times (in days) at which the fluxes were predicted.
        last_fluxes : numpy.ndarray
            A length T1 x W array of the last valid flux values at each wavelength
            at the last valid times (in nJy).
        query_times : numpy.ndarray
            A length T2 array of the query times (in days) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A T2 x W matrix of extrapolated values.
        """
        last_fluxes = last_fluxes.T
        return self._extrapolate(last_times, last_fluxes, query_times).T

    def extrapolate_wavelength(self, last_waves, last_fluxes, query_waves):
        """Extrapolate along the wavelength axis.

        Parameters
        ----------
        last_waves : float or np.ndarray
            A length W1 array of the last valid wavelengths (in AA) at which the fluxes were predicted.
        last_fluxes : numpy.ndarray
            A length T x W1 array of the last valid flux values at each time
            at the last valid wavelength (in nJy).
        query_waves : numpy.ndarray
            A length W2 array of the query wavelengths (in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A T x W2 matrix of extrapolated values.
        """
        # We transpose the result to turn the W2 x T matrix into a T x W matrix.
        return self._extrapolate(last_waves, last_fluxes, query_waves)


class ZeroPadding(FluxExtrapolationModel):
    """Extrapolate by zero padding the results."""

    def __init__(self):
        super().__init__()

    def _extrapolate(self, last_values, last_fluxes, query_values):
        """Evaluate the extrapolation given the last valid points(s) and a list of new
        query points.

        Note
        ----
        This function does not care which axis is being extrapolated. The returned values are
        always len(query_values) x len(last_fluxes) and may need to be transposed by the calling
        function.

        Parameters
        ----------
        last_values : float or np.ndarray
            The last valid value along the extrapolation axis at which the flux was predicted
            (e.g., wavelength in AA or time in days).
        last_fluxes : numpy.ndarray
            A length N array of the flux values at the last valid time or wavelength (in nJy).
        query_values : numpy.ndarray
            A length M array of values along the extrapolation axis (times in days or wavelengths
            in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A N x M matrix of extrapolated values. Where M is the number of query points and
            N is the number of flux values at the last valid point.
        """

        N_len = len(last_fluxes)
        M_len = len(query_values)
        return np.zeros((N_len, M_len))


class ConstantPadding(FluxExtrapolationModel):
    """Extrapolate using a constant value in nJy.

    Attributes
    ----------
    value : float
        The value (in nJy) to use for the extrapolation.
    """

    def __init__(self, value=0.0):
        super().__init__()

        if value < 0:
            raise ValueError("Extrapolation value must be positive.")
        self.value = value

    def _extrapolate(self, last_values, last_fluxes, query_values):
        """Evaluate the extrapolation given the last valid points(s) and a list of new
        query points.

        Note
        ----
        This function does not care which axis is being extrapolated. The returned values are
        always len(query_values) x len(last_fluxes) and may need to be transposed by the calling
        function.

        Parameters
        ----------
        last_values : float or np.ndarray
            The last valid value along the extrapolation axis at which the flux was predicted
            (e.g., wavelength in AA or time in days).
        last_fluxes : numpy.ndarray
            A length N array of the flux values at the last valid time or wavelength (in nJy).
        query_values : numpy.ndarray
            A length M array of values along the extrapolation axis (times in days or wavelengths
            in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A N x M matrix of extrapolated values. Where M is the number of query points and
            N is the number of flux values at the last valid point.
        """

        N_len = len(last_fluxes)
        M_len = len(query_values)
        return np.full((N_len, M_len), self.value)


class LastValue(FluxExtrapolationModel):
    """Extrapolate using the last valid value along the desired axis."""

    def __init__(self):
        super().__init__()

    def _extrapolate(self, last_values, last_fluxes, query_values):
        """Evaluate the extrapolation given the last valid points(s) and a list of new
        query points.

        Note
        ----
        This function does not care which axis is being extrapolated. The returned values are
        always len(query_values) x len(last_fluxes) and may need to be transposed by the calling
        function.

        Parameters
        ----------
        last_values : float or np.ndarray
            The last valid value along the extrapolation axis at which the flux was predicted
            (e.g., wavelength in AA or time in days).
        last_fluxes : numpy.ndarray
            A length N array of the flux values at the last valid time or wavelength (in nJy).
        query_values : numpy.ndarray
            A length M array of values along the extrapolation axis (times in days or wavelengths
            in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A N x M matrix of extrapolated values. Where M is the number of query points and
            N is the number of flux values at the last valid point.
        """

        last_fluxes = np.asarray(last_fluxes).reshape(-1)

        return np.tile(last_fluxes[:, np.newaxis], (1, len(query_values)))


class LinearDecay(FluxExtrapolationModel):
    """Linear decay of the flux using the last valid point(s) down to zero.

    Attributes
    ----------
    decay_width : float or np.ndarray
        The width of the decay region in Angstroms. The flux is
        linearly decreased to zero over this range.
    """

    def __init__(self, decay_width=100.0):
        super().__init__()

        if decay_width <= 0:
            raise ValueError("decay_width must be positive.")
        self.decay_width = decay_width

    def _extrapolate(self, last_values, last_fluxes, query_values):
        """Evaluate the extrapolation given the last valid points(s) and a list of new
        query points.

        Note
        ----
        This function does not care which axis is being extrapolated. The returned values are
        always len(query_values) x len(last_fluxes) and may need to be transposed by the calling
        function.

        Parameters
        ----------
        last_values : float or np.ndarray
            The last valid value along the extrapolation axis at which the flux was predicted
            (e.g., wavelength in AA or time in days).
        last_fluxes : numpy.ndarray
            A length N array of the flux values at the last valid time or wavelength (in nJy).
        query_values : numpy.ndarray
            A length M array of values along the extrapolation axis (times in days or wavelengths
            in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A N x M matrix of extrapolated values. Where M is the number of query points and
            N is the number of flux values at the last valid point.
        """

        last_fluxes = np.asarray(last_fluxes).reshape(-1)
        query_values = np.asarray(query_values)
        dist = np.abs(query_values - last_values)

        multiplier = np.clip(1.0 - (dist / self.decay_width), 0.0, 1.0)

        flux = last_fluxes[:, np.newaxis] * multiplier[np.newaxis, :]

        return flux


class LinearDecayOnMag(FluxExtrapolationModel):
    """Linear decay of the converted magnitude using the last valid point(s) with a fixed decay
    rate down to a specific magnitude threshold. This is generally used for extrapolating in
    the time axis.

    Attributes
    ----------
    decay_rate : float or np.ndarray
        The rate of the decay region in days. The magnitude is
        linearly decreased to the mag_thres over this range.
    mag_thres : float or np.ndarray
        The mag threshold for the linear decay extraplation. Fluxes are capped at this value
        for time/wavelength beyond this value.
    """

    def __init__(self, decay_rate=0.02, mag_thres=40.0):
        super().__init__()

        if decay_rate <= 0:
            raise ValueError("decay_rate must be positive.")
        self.decay_rate = decay_rate
        self.mag_thres = mag_thres

    def _extrapolate(self, last_values, last_fluxes, query_values):
        """Evaluate the extrapolation given the last valid points(s) and a list of new
        query points.

        Note
        ----
        This function does not care which axis is being extrapolated. The returned values are
        always len(query_values) x len(last_fluxes) and may need to be transposed by the calling
        function.

        Parameters
        ----------
        last_values : float or np.ndarray
            The last valid value along the extrapolation axis at which the flux was predicted
            (e.g., wavelength in AA or time in days).
        last_fluxes : numpy.ndarray
            A length N array of the flux values at the last valid time or wavelength (in nJy).
        query_values : numpy.ndarray
            A length M array of values along the extrapolation axis (times in days or wavelengths
            in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A N x M matrix of extrapolated values. Where M is the number of query points and
            N is the number of flux values at the last valid point.
        """

        last_fluxes = np.asarray(last_fluxes).reshape(-1)
        last_fluxes = np.clip(last_fluxes, mag2flux(self.mag_thres), None)
        last_mags = flux2mag(last_fluxes)

        query_values = np.asarray(query_values)

        dist = np.abs(query_values - last_values)

        mag = np.clip(last_mags[:, np.newaxis] + dist[np.newaxis, :] * self.decay_rate, None, self.mag_thres)

        return mag2flux(mag)


class ExponentialDecay(FluxExtrapolationModel):
    """Exponential decay of the flux using the last valid point(s) down to zero.

    f(t, λ) = f(t, λ_last) * exp(- rate * \\|λ - λ_last\\|)

    Attributes
    ----------
    rate : float
        The decay rate in the exponential function.
    """

    def __init__(self, rate):
        super().__init__()

        if rate < 0:
            raise ValueError("Decay rate must be zero or positive.")
        self.rate = rate

    def _extrapolate(self, last_values, last_fluxes, query_values):
        """Evaluate the extrapolation given the last valid points(s) and a list of new
        query points.

        Note
        ----
        This function does not care which axis is being extrapolated. The returned values are
        always len(query_values) x len(last_fluxes) and may need to be transposed by the calling
        function.

        Parameters
        ----------
        last_values : float or np.ndarray
            The last valid value along the extrapolation axis at which the flux was predicted
            (e.g., wavelength in AA or time in days).
        last_fluxes : numpy.ndarray
            A length N array of the flux values at the last valid time or wavelength (in nJy).
        query_values : numpy.ndarray
            A length M array of values along the extrapolation axis (times in days or wavelengths
            in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A N x M matrix of extrapolated values. Where M is the number of query points and
            N is the number of flux values at the last valid point.
        """

        last_fluxes = np.asarray(last_fluxes).reshape(-1)
        query_values = np.asarray(query_values)
        dist = np.abs(query_values - last_values)

        multiplier = np.exp(-self.rate * dist)
        flux = last_fluxes[:, np.newaxis] * multiplier[np.newaxis, :]
        return flux


def _bin_rows_median(last_fluxes, nbin, *, nan_safe=True):
    """Bin the input fluxes on the first axis given number of bins and return the median values
       of each bin. This is used for binning the last fluxes to avoid extrapolating to extreme
       values.

    Parameters
    ----------
    last_fluxes : np.ndarray
        A N x T array of the input fluxes to be binned.
    nbin : int
        Number of bins along the first axis.
    nan_safe : bool, optional
        If True, use np.nanmedian (ignore NaNs).
        If False, use np.median.

    Returns
    -------
    binned_fluxes : np.ndarray
        A nbin x T array of the binned fluxes.
    """
    last_fluxes = np.asarray(last_fluxes)
    N, T = last_fluxes.shape

    if nbin > N:
        raise ValueError("nbin must be smaller or equal to N")
    # Bin edges that evenly partition rows
    edges = np.linspace(0, N, nbin + 1, dtype=int)

    binned_fluxes = np.empty((nbin, T), dtype=float)
    for b in range(nbin):
        lo, hi = edges[b], edges[b + 1]
        chunk = last_fluxes[lo:hi]

        binned_fluxes[b] = np.nanmedian(chunk, axis=0) if nan_safe else np.median(chunk, axis=0)

    return binned_fluxes


class LinearFit(FluxExtrapolationModel):
    """Linear extrapolation based on a linear fit to the last few points.

    Parameters
    ----------
    nfit : int
        The number of points to be used for extrapolation. (Default is 5)
    nbin : int
        The number of bins to be used to bin the last fluxes. This can be used to avoid extrapolating
        to extreme values when models are not well-behaved in smaller bins.
    """

    def __init__(self, nfit=5, nbin=None):
        super().__init__()
        self.nfit = nfit
        self.nbin = nbin

    def _extrapolate(self, last_values, last_fluxes, query_values):
        """Evaluate the extrapolation given the last valid points(s) and a list of new
        query points.

        Parameters
        ----------
        last_values : np.ndarray
            A T elements array of the last values along the extrapolation axis at which the flux was predicted
            (e.g., wavelength in AA or time in days).
        last_fluxes : ndarray
            A length N x T matrix of the flux values at the last valid time or wavelength (in nJy).
        query_values : ndarray
            A length M array of values along the extrapolation axis (times in days or wavelengths
            in AA) at which to extrapolate.

        Returns
        -------
        flux : ndarray
            A N x M matrix of extrapolated values. Where M is the number of query points and
            N is the number of flux values at the last valid point.
        """

        if len(last_values) <= 1:
            raise ValueError("Need at least two points to extrapolate using this method.")

        N = last_fluxes.shape[0]

        if self.nbin is None:
            binned_fluxes = last_fluxes
        else:
            # guard: can't have more bins than rows
            nbin = int(min(self.nbin, N))
            binned_fluxes = _bin_rows_median(last_fluxes, nbin=nbin, nan_safe=True)

        A = np.column_stack([last_values, np.ones_like(last_values)])
        B = np.array(binned_fluxes, dtype=float, copy=True).T

        coeffs = np.linalg.lstsq(A, B, rcond=None)[0]
        slope, intercept = coeffs

        # (nbin, M)
        flux_binned = slope[:, None] * query_values[None, :] + intercept[:, None]
        flux_binned = np.clip(flux_binned, 0.0, None)

        # Expand back to (N, M): row i gets its bin's curve
        if self.nbin is None:
            flux = flux_binned
        else:
            row_to_bin = (np.arange(N) * nbin) // N
            flux = flux_binned[row_to_bin]

        return flux


class LinearFitOnMag(FluxExtrapolationModel):
    """Linear extrapolation based on a linear fit to the coverted magnitude of the last few points.

    Parameters
    ----------
    nfit : int
        The number of points to be used for extrapolation. (Default is 5)
    nbin : int
        The number of bins to be used to bin the last fluxes. This can be used to avoid extrapolating
        to extreme values when models are not well-behaved in smaller bins.
    """

    def __init__(self, nfit=5, nbin=None):
        super().__init__()
        self.nfit = nfit
        self.nbin = nbin

    def _extrapolate(self, last_values, last_fluxes, query_values):
        """Evaluate the extrapolation given the last valid points(s) and a list of new
        query points.

        Parameters
        ----------
        last_values : np.ndarray
            A T elements array of the last values along the extrapolation axis at which the flux was predicted
            (e.g., wavelength in AA or time in days).
        last_fluxes : ndarray
            A length N x T matrix of the flux values at the last valid time or wavelength (in nJy).
        query_values : ndarray
            A length M array of values along the extrapolation axis (times in days or wavelengths
            in AA) at which to extrapolate.

        Returns
        -------
        flux : ndarray
            A N x M matrix of extrapolated values. Where M is the number of query points and
            N is the number of flux values at the last valid point.
        """

        if len(last_values) <= 1:
            raise ValueError("Need at least two points to extrapolate using this method.")

        N = last_fluxes.shape[0]

        last_fluxes = np.clip(last_fluxes, 1.0e-40, None)
        last_fluxes = flux2mag(last_fluxes)

        if self.nbin is None:
            binned_fluxes = last_fluxes
        else:
            # guard: can't have more bins than rows (otherwise some bins empty -> median NaN)
            nbin = int(min(self.nbin, N))
            binned_fluxes = _bin_rows_median(last_fluxes, nbin=nbin, nan_safe=True)

        A = np.column_stack([last_values, np.ones_like(last_values)])
        B = np.array(binned_fluxes, dtype=float, copy=True).T

        coeffs = np.linalg.lstsq(A, B, rcond=None)[0]
        slope, intercept = coeffs

        # (nbin, M)
        flux_binned = slope[:, None] * query_values[None, :] + intercept[:, None]
        flux_binned = np.clip(flux_binned, 0.0, None)

        if self.nbin is None:
            flux = flux_binned
        else:
            # Expand back to (N, M): row i gets its bin's curve
            row_to_bin = (np.arange(N) * nbin) // N
            flux = flux_binned[row_to_bin]

        return mag2flux(flux)



import numpy as np
import pytest
from lightcurvelynx.astro_utils.mag_flux import mag2flux
from lightcurvelynx.utils.extrapolate import (
    ConstantPadding,
    ExponentialDecay,
    LastValue,
    LinearDecay,
    LinearDecayOnMag,
    LinearFit,
    LinearFitOnMag,
    ZeroPadding,
)


def test_flux_extrapolation_model():
    """Test the base class for the extrapolation methods."""
    # Create an instance of the base class
    extrapolator = ZeroPadding()

    # Test that extrapolating along wavelength returns a zero matrix
    last_wave = 1000.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1200.0])
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    expected_result = np.zeros((3, 4))
    np.testing.assert_allclose(result, expected_result)

    # Test that we get a correctly shaped 2-D array back even if we have
    # only one query point.
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, np.array([1025.0]))
    assert result.shape == (3, 1)

    # Test that extrapolating along time returns a zero matrix
    last_time = 0.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_times = np.array([0.0, 1.0, 2.0, 5.0])
    result = extrapolator.extrapolate_time(last_time, last_flux, query_times)
    expected_result = np.zeros((4, 3))
    np.testing.assert_allclose(result, expected_result)

    # Test that we get a correctly shaped 2-D array back even if we have
    # only one query point.
    result = extrapolator.extrapolate_time(last_time, last_flux, np.array([1.0]))
    assert result.shape == (1, 3)


def test_constant_extrapolation():
    """Test that the constant extrapolation function works."""
    # Create an instance of the ConstantPadding class
    extrapolator = ConstantPadding(value=100.0)

    # Test extrapolation past the last valid point.
    last_wave = 1000.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1200.0])
    expected_flux = np.full((3, 4), 100.0)
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test that extrapolating along time returns a zero matrix
    last_time = 0.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_times = np.array([0.0, 1.0, 2.0, 5.0])
    result = extrapolator.extrapolate_time(last_time, last_flux, query_times)
    expected_result = np.full((4, 3), 100.0)
    np.testing.assert_allclose(result, expected_result)

    # Test that we fail if the value is not positive
    with pytest.raises(ValueError):
        _ = ConstantPadding(value=-1)


def test_last_value_extrapolation():
    """Test that the last value extrapolation function works."""
    # Create an instance of the LastValue class
    extrapolator = LastValue()

    # Test extrapolation past the last valid point.
    last_wave = 1000.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1200.0])
    expected_flux = np.array(
        [
            [100.0, 100.0, 100.0, 100.0],
            [200.0, 200.0, 200.0, 200.0],
            [300.0, 300.0, 300.0, 300.0],
        ]
    )
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test extrapolation along time.
    first_time = 0.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_times = np.array([-2.0, -1.0])
    expected_flux = np.array(
        [
            [100.0, 200.0, 300.0],
            [100.0, 200.0, 300.0],
        ]
    )
    result = extrapolator.extrapolate_time(first_time, last_flux, query_times)
    np.testing.assert_allclose(result, expected_flux)


def test_linear_decay_extrapolate():
    """Test that the linear decay function works."""
    decay_width = 100.0
    extrapolator = LinearDecay(decay_width=decay_width)

    # Test extrapolation past the last valid point.
    last_wave = 1000.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1075.0, 1100.0, 1150.0])
    expected_flux = np.array(
        [
            [100.0, 75.0, 50.0, 25.0, 0.0, 0.0],
            [200.0, 150.0, 100.0, 50.0, 0.0, 0.0],
            [300.0, 225.0, 150.0, 75.0, 0.0, 0.0],
        ]
    )
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test extrapolation before the first valid point.
    query_waves = np.array([1000.0, 975.0, 950.0, 925.0, 900.0, 850.0])
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test that we fail if the decay width is not positive
    with pytest.raises(ValueError):
        _ = LinearDecay(decay_width=-1.0)


def test_linear_decay_on_mag_extrapolate():
    """Test that the linear decay function works."""
    decay_rate = 0.1
    extrapolator = LinearDecayOnMag(decay_rate=decay_rate, mag_thres=20.0)

    # Test extrapolation past the last valid point.
    last_wave = 1000.0
    last_mag = np.array([10.0, 11.0, 12.0])
    last_flux = mag2flux(last_mag)
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1075.0, 1100.0, 1150.0])
    expected_mag = np.array(
        [
            [10.0, 12.5, 15.0, 17.5, 20.0, 20.0],
            [11.0, 13.5, 16.0, 18.5, 20.0, 20.0],
            [12.0, 14.5, 17.0, 19.5, 20.0, 20.0],
        ]
    )
    expected_flux = mag2flux(expected_mag)
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test extrapolation before the first valid point.
    query_waves = np.array([1000.0, 975.0, 950.0, 925.0, 900.0, 850.0])
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test that we fail if the decay rate is not positive
    with pytest.raises(ValueError):
        _ = LinearDecayOnMag(decay_rate=-1.0)


def test_exponential_decay_extrapolate():
    """Test that the exponential decay function works."""
    extrapolator = ExponentialDecay(rate=0.1)

    # Test extrapolation past the last valid point.
    last_wave = 1000.0
    last_flux = np.array([100.0, 200.0, 300.0])
    query_waves = np.array([1000.0, 1025.0, 1050.0, 1075.0, 1100.0, 1150.0])

    t0_flux = 100.0 * np.exp([-0.0, -2.5, -5.0, -7.5, -10.0, -15.0])
    expected_flux = np.vstack((t0_flux, t0_flux * 2, t0_flux * 3))
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test extrapolation before the first valid point.
    query_waves = np.array([1000.0, 975.0, 950.0, 925.0, 900.0, 850.0])
    result = extrapolator.extrapolate_wavelength(last_wave, last_flux, query_waves)
    np.testing.assert_allclose(result, expected_flux)

    # Test that we fail if the decay rate is not positive
    with pytest.raises(ValueError):
        _ = ExponentialDecay(rate=-0.1)


def test_linear_fit_extrapolate():
    """Test that the linear fit extrapolation works."""

    extrapolator = LinearFit()

    last_times = np.array([30.0, 40.0, 50.0, 55.0])
    last_fluxes = np.repeat([[300.0], [250.0], [200.0], [175]], 5, axis=1)
    query_times = np.array([60.0, 70.0, 80.0])
    expected_flux = np.repeat([[150.0], [100.0], [50.0]], 5, axis=1)
    result = extrapolator.extrapolate_time(last_times, last_fluxes, query_times)
    np.testing.assert_allclose(result, expected_flux)

    last_waves = np.array([30.0, 40.0, 50.0, 55.0])
    last_fluxes = np.repeat([[300.0], [250.0], [200.0], [175]], 5, axis=1).T
    query_waves = np.array([60.0, 70.0, 80.0])
    expected_flux = np.repeat([[150.0], [100.0], [50.0]], 5, axis=1).T
    result = extrapolator.extrapolate_wavelength(last_waves, last_fluxes, query_waves)
    np.testing.assert_allclose(result, expected_flux)


def test_linear_fit_on_mag_extrapolate():
    """Test that the linear fit extrapolation works."""

    extrapolator = LinearFitOnMag()

    last_times = np.array([30.0, 40.0, 50.0, 55.0])
    last_fluxes = mag2flux(np.repeat([[21.0], [21.2], [21.4], [21.5]], 5, axis=1))
    query_times = np.array([60.0, 70.0, 80.0])
    expected_flux = mag2flux(np.repeat([[21.6], [21.8], [22.0]], 5, axis=1))
    result = extrapolator.extrapolate_time(last_times, last_fluxes, query_times)
    np.testing.assert_allclose(result, expected_flux)

    last_waves = np.array([30.0, 40.0, 50.0, 55.0])
    last_fluxes = mag2flux(np.repeat([[21.0], [21.2], [21.4], [21.5]], 5, axis=1)).T
    query_waves = np.array([60.0, 70.0, 80.0])
    expected_flux = mag2flux(np.repeat([[21.6], [21.8], [22.0]], 5, axis=1)).T
    result = extrapolator.extrapolate_wavelength(last_waves, last_fluxes, query_waves)
    np.testing.assert_allclose(result, expected_flux)


def test_linear_fit_extrapolate_binning():
    """Test that binning works for LinearFit"""

    extrapolator = LinearFit(nbin=1)
    assert extrapolator.nbin == 1

    last_times = np.array([30.0, 40.0, 50.0, 55.0])
    last_fluxes = np.repeat(
        [[300.0, 305.0, 295.0], [250.0, 255.0, 245.0], [200.0, 205.0, 195.0], [175.0, 170.0, 180.0]],
        5,
        axis=1,
    )
    query_times = np.array([60.0, 70.0, 80.0])
    expected_flux = np.repeat([[150.0], [100.0], [50.0]], 15, axis=1)
    result = extrapolator.extrapolate_time(last_times, last_fluxes, query_times)
    np.testing.assert_allclose(result, expected_flux)

    extrapolator = LinearFit(nbin=20)
    last_times = np.array([30.0, 40.0, 50.0, 55.0])
    last_fluxes = np.repeat(
        [[300.0, 305.0, 295.0], [250.0, 255.0, 245.0], [200.0, 205.0, 195.0], [175.0, 180.0, 170.0]],
        5,
        axis=1,
    )
    query_times = np.array([60.0, 70.0, 80.0])
    expected_flux = np.repeat([[150.0, 155.0, 145.0], [100.0, 105.0, 95.0], [50.0, 55.0, 45.0]], 5, axis=1)
    result = extrapolator.extrapolate_time(last_times, last_fluxes, query_times)
    np.testing.assert_allclose(result, expected_flux)


def test_linear_fit_on_mag_extrapolate_binning():
    """Test that binning works for LinearFitOnMag"""

    extrapolator = LinearFitOnMag(nbin=1)
    assert extrapolator.nbin == 1

    last_times = np.array([30.0, 40.0, 50.0, 55.0])
    last_fluxes = mag2flux(
        np.repeat(
            [[21.0, 21.05, 20.95], [21.2, 21.15, 21.25], [21.4, 21.35, 21.45], [21.5, 21.45, 21.55]],
            5,
            axis=1,
        )
    )
    query_times = np.array([60.0, 70.0, 80.0])
    expected_flux = mag2flux(np.repeat([[21.6], [21.8], [22.0]], 15, axis=1))
    result = extrapolator.extrapolate_time(last_times, last_fluxes, query_times)
    np.testing.assert_allclose(result, expected_flux)

    extrapolator = LinearFitOnMag(nbin=20)
    last_times = np.array([30.0, 40.0, 50.0, 55.0])
    last_fluxes = mag2flux(
        np.repeat(
            [[21.0, 20.95, 21.05], [21.2, 21.15, 21.25], [21.4, 21.35, 21.45], [21.5, 21.45, 21.55]],
            5,
            axis=1,
        )
    )
    query_times = np.array([60.0, 70.0, 80.0])
    expected_flux = mag2flux(
        np.repeat([[21.6, 21.55, 21.65], [21.8, 21.75, 21.85], [22.0, 21.95, 22.05]], 5, axis=1)
    )
    result = extrapolator.extrapolate_time(last_times, last_fluxes, query_times)
    np.testing.assert_allclose(result, expected_flux)
